# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from testing_support.fixtures import override_application_settings
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.validators.validate_transaction_errors import validate_transaction_errors

from testing_support.validators.validate_code_level_metrics import validate_code_level_metrics

def target_application():
    # We need to delay Flask application creation because of ordering
    # issues whereby the agent needs to be initialised before Flask is
    # imported and the routes configured. Normally pytest only runs the
    # global fixture which will initialise the agent after each test
    # file is imported, which is too late. We also can't do application
    # creation within a function as we will then get view handler
    # functions are different between Python 2 and 3, with the latter
    # showing <local> scope in path.

    from _test_compress import _test_application
    return _test_application


_test_application_index_scoped_metrics = [
        ('Function/flask.app:Flask.wsgi_app', 1),
        ('Python/WSGI/Application', 1),
        ('Python/WSGI/Response', 1),
        ('Python/WSGI/Finalize', 1),
        ('Function/_test_compress:index_page', 1),
        ('Function/werkzeug.wsgi:ClosingIterator.close', 1)]


@validate_code_level_metrics("_test_compress", "index_page")
@validate_transaction_errors(errors=[])
@validate_transaction_metrics('_test_compress:index_page',
        scoped_metrics=_test_application_index_scoped_metrics)
def test_compress_middleware():
    application = target_application()
    response = application.get('/compress')
    response.mustcontain(500 * 'X')


_test_html_insertion_settings = {
    'browser_monitoring.enabled': True,
    'browser_monitoring.auto_instrument': True,
    'browser_monitoring.content_type': ['text/html'],
    'js_agent_loader': u'<!-- NREUM HEADER -->',
}


@override_application_settings(_test_html_insertion_settings)
def test_html_insertion_flask_middleware():
    application = target_application()
    headers = {'Accept-Encoding': 'gzip'}
    response = application.get('/html_insertion', headers=headers, status=200)

    # The 'NREUM HEADER' value comes from our override for the header.
    # The 'NREUM.info' value comes from the programmatically generated
    # footer added by the agent.

    response.mustcontain('NREUM HEADER', 'NREUM.info')


@override_application_settings(_test_html_insertion_settings)
def test_html_inserted_for_html_served_from_file():
    application = target_application()
    headers = {'Accept-Encoding': 'gzip'}
    response = application.get('/html_served_from_file',
            headers=headers, status=200)

    # The 'NREUM HEADER' value comes from our override for the header.
    # The 'NREUM.info' value comes from the programmatically generated
    # footer added by the agent.

    response.mustcontain('NREUM HEADER', 'NREUM.info')


_test_html_insertion_manual_settings = {
    'browser_monitoring.enabled': True,
    'browser_monitoring.auto_instrument': True,
    'browser_monitoring.content_type': ['text/html'],
    'js_agent_loader': u'<!-- NREUM HEADER -->',
}


@override_application_settings(_test_html_insertion_manual_settings)
def test_html_insertion_manual_flask_middleware():
    application = target_application()
    response = application.get('/html_insertion_manual', status=200)

    # The 'NREUM HEADER' value comes from our override for the header.
    # The 'NREUM.info' value comes from the programmatically generated
    # footer added by the agent.

    response.mustcontain(no=['NREUM HEADER', 'NREUM.info'])


@override_application_settings(_test_html_insertion_settings)
def test_html_insertion_unnamed_attachment_header_flask_middleware():
    application = target_application()
    response = application.get(
            '/html_insertion_unnamed_attachment_header', status=200)

    # The 'NREUM HEADER' value comes from our override for the header.
    # The 'NREUM.info' value comes from the programmatically generated
    # footer added by the agent.

    response.mustcontain(no=['NREUM HEADER', 'NREUM.info'])


@override_application_settings(_test_html_insertion_settings)
def test_html_insertion_named_attachment_header_flask_middleware():
    application = target_application()
    response = application.get(
            '/html_insertion_named_attachment_header', status=200)

    # The 'NREUM HEADER' value comes from our override for the header.
    # The 'NREUM.info' value comes from the programmatically generated
    # footer added by the agent.

    response.mustcontain(no=['NREUM HEADER', 'NREUM.info'])


@override_application_settings(_test_html_insertion_settings)
def test_text_served_from_file():
    application = target_application()
    response = application.get(
            '/text_served_from_file', status=200)

    # The 'NREUM HEADER' value comes from our override for the header.
    # The 'NREUM.info' value comes from the programmatically generated
    # footer added by the agent.

    response.mustcontain(no=['NREUM HEADER', 'NREUM.info'])


@override_application_settings(_test_html_insertion_settings)
def test_empty_content_type():
    application = target_application()
    response = application.get('/empty_content_type')

    # Make sure agent can handle content type of ''

    assert response.headers['Content-Type'] == ''
    response.mustcontain(no=['NREUM HEADER', 'NREUM.info'])



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

import numpy as np
import pandas
from testing_support.fixtures import (
    override_application_settings,
    reset_core_stats_engine,
)
from testing_support.fixtures import override_application_settings
from testing_support.validators.validate_ml_event_count import validate_ml_event_count
from testing_support.validators.validate_ml_events import validate_ml_events

from newrelic.api.background_task import background_task

pandas_df_category_recorded_custom_events = [
    (
        {"type": "InferenceData"},
        {
            "inference_id": None,
            "prediction_id": None,
            "modelName": "DecisionTreeClassifier",
            "model_version": "0.0.0",
            "feature.col1": 2.0,
            "feature.col2": 4.0,
            "label.0": "27.0",
            "new_relic_data_schema_version": 2,
        },
    ),
]


@reset_core_stats_engine()
def test_pandas_df_categorical_feature_event():
    @validate_ml_events(pandas_df_category_recorded_custom_events)
    @validate_ml_event_count(count=1)
    @background_task()
    def _test():
        import sklearn.tree

        clf = getattr(sklearn.tree, "DecisionTreeClassifier")(random_state=0)
        model = clf.fit(
            pandas.DataFrame({"col1": [27.0, 24.0], "col2": [23.0, 25.0]}, dtype="category"),
            pandas.DataFrame({"label": [27.0, 28.0]}),
        )

        labels = model.predict(pandas.DataFrame({"col1": [2.0], "col2": [4.0]}, dtype="category"))
        return model

    _test()


label_type = "bool" if sys.version_info < (3, 8) else "numeric"
true_label_value = "True" if sys.version_info < (3, 8) else "1.0"
false_label_value = "False" if sys.version_info < (3, 8) else "0.0"
pandas_df_bool_recorded_custom_events = [
    (
        {"type": "InferenceData"},
        {
            "inference_id": None,
            "prediction_id": None,
            "modelName": "DecisionTreeClassifier",
            "model_version": "0.0.0",
            "feature.col1": True,
            "feature.col2": True,
            "label.0": true_label_value,
            "new_relic_data_schema_version": 2,
        },
    ),
]


@reset_core_stats_engine()
def test_pandas_df_bool_feature_event():
    @validate_ml_events(pandas_df_bool_recorded_custom_events)
    @validate_ml_event_count(count=1)
    @background_task()
    def _test():
        import sklearn.tree

        dtype_name = "bool" if sys.version_info < (3, 8) else "boolean"
        x_train = pandas.DataFrame({"col1": [True, False], "col2": [True, False]}, dtype=dtype_name)
        y_train = pandas.DataFrame({"label": [True, False]}, dtype=dtype_name)
        x_test = pandas.DataFrame({"col1": [True], "col2": [True]}, dtype=dtype_name)

        clf = getattr(sklearn.tree, "DecisionTreeClassifier")(random_state=0)
        model = clf.fit(x_train, y_train)

        labels = model.predict(x_test)
        return model

    _test()


pandas_df_float_recorded_custom_events = [
    (
        {"type": "InferenceData"},
        {
            "inference_id": None,
            "prediction_id": None,
            "modelName": "DecisionTreeRegressor",
            "model_version": "0.0.0",
            "feature.col1": 100.0,
            "feature.col2": 300.0,
            "label.0": "345.6",
            "new_relic_data_schema_version": 2,
        },
    ),
]


@reset_core_stats_engine()
def test_pandas_df_float_feature_event():
    @validate_ml_events(pandas_df_float_recorded_custom_events)
    @validate_ml_event_count(count=1)
    @background_task()
    def _test():
        import sklearn.tree

        x_train = pandas.DataFrame({"col1": [120.0, 254.0], "col2": [236.9, 234.5]}, dtype="float64")
        y_train = pandas.DataFrame({"label": [345.6, 456.7]}, dtype="float64")
        x_test = pandas.DataFrame({"col1": [100.0], "col2": [300.0]}, dtype="float64")

        clf = getattr(sklearn.tree, "DecisionTreeRegressor")(random_state=0)

        model = clf.fit(x_train, y_train)
        labels = model.predict(x_test)

        return model

    _test()


int_list_recorded_custom_events = [
    (
        {"type": "InferenceData"},
        {
            "inference_id": None,
            "prediction_id": None,
            "modelName": "ExtraTreeRegressor",
            "model_version": "0.0.0",
            "feature.0": 1,
            "feature.1": 2,
            "label.0": "1.0",
            "new_relic_data_schema_version": 2,
        },
    ),
]


@reset_core_stats_engine()
def test_int_list():
    @validate_ml_events(int_list_recorded_custom_events)
    @validate_ml_event_count(count=1)
    @background_task()
    def _test():
        import sklearn.tree

        x_train = [[0, 0], [1, 1]]
        y_train = [0, 1]
        x_test = [[1, 2]]

        clf = getattr(sklearn.tree, "ExtraTreeRegressor")(random_state=0)
        model = clf.fit(x_train, y_train)

        labels = model.predict(x_test)
        return model

    _test()


numpy_int_recorded_custom_events = [
    (
        {"type": "InferenceData"},
        {
            "inference_id": None,
            "prediction_id": None,
            "modelName": "ExtraTreeRegressor",
            "model_version": "0.0.0",
            "feature.0": 12,
            "feature.1": 13,
            "label.0": "11.0",
            "new_relic_data_schema_version": 2,
        },
    ),
]


@reset_core_stats_engine()
def test_numpy_int_array():
    @validate_ml_events(numpy_int_recorded_custom_events)
    @validate_ml_event_count(count=1)
    @background_task()
    def _test():
        import sklearn.tree

        x_train = np.array([[10, 10], [11, 11]], dtype="int")
        y_train = np.array([10, 11], dtype="int")
        x_test = np.array([[12, 13]], dtype="int")

        clf = getattr(sklearn.tree, "ExtraTreeRegressor")(random_state=0)
        model = clf.fit(x_train, y_train)

        labels = model.predict(x_test)
        return model

    _test()


numpy_str_recorded_custom_events = [
    (
        {"type": "InferenceData"},
        {
            "inference_id": None,
            "prediction_id": None,
            "modelName": "DecisionTreeClassifier",
            "model_version": "0.0.0",
            "feature.0": "20",
            "feature.1": "21",
            "label.0": "21",
            "new_relic_data_schema_version": 2,
        },
    ),
    (
        {"type": "InferenceData"},
        {
            "inference_id": None,
            "prediction_id": None,
            "modelName": "DecisionTreeClassifier",
            "model_version": "0.0.0",
            "feature.0": "22",
            "feature.1": "23",
            "label.0": "21",
            "new_relic_data_schema_version": 2,
        },
    ),
]


@reset_core_stats_engine()
def test_numpy_str_array_multiple_features():
    @validate_ml_events(numpy_str_recorded_custom_events)
    @validate_ml_event_count(count=2)
    @background_task()
    def _test():
        import sklearn.tree

        x_train = np.array([[20, 20], [21, 21]], dtype="<U4")
        y_train = np.array([20, 21], dtype="<U4")
        x_test = np.array([[20, 21], [22, 23]], dtype="<U4")
        clf = getattr(sklearn.tree, "DecisionTreeClassifier")(random_state=0)

        model = clf.fit(x_train, y_train)
        labels = model.predict(x_test)

        return model

    _test()


numpy_str_recorded_custom_events_no_value = [
    (
        {"type": "InferenceData"},
        {
            "inference_id": None,
            "prediction_id": None,
            "modelName": "DecisionTreeClassifier",
            "model_version": "0.0.0",
            "new_relic_data_schema_version": 2,
        },
    ),
]


disabled_inference_value_settings = {
    "machine_learning.enabled": True,
    "machine_learning.inference_events_value.enabled": False,
    "ml_insights_events.enabled": True
}


@override_application_settings(disabled_inference_value_settings)
@reset_core_stats_engine()
def test_does_not_include_value_when_inference_event_value_enabled_is_false():
    @validate_ml_events(numpy_str_recorded_custom_events_no_value)
    @validate_ml_event_count(count=1)
    @background_task()
    def _test():
        import sklearn.tree

        x_train = np.array([[20, 20], [21, 21]], dtype="<U4")
        y_train = np.array([20, 21], dtype="<U4")
        x_test = np.array([[20, 21]], dtype="<U4")
        clf = getattr(sklearn.tree, "DecisionTreeClassifier")(random_state=0)

        model = clf.fit(x_train, y_train)
        labels = model.predict(x_test)

        return model

    _test()


disabled_ml_insights_settings = {
    "machine_learning.enabled": True,
    "machine_learning.inference_events_value.enabled": True,
    "ml_insights_events.enabled": False
}


@override_application_settings(disabled_ml_insights_settings)
@reset_core_stats_engine()
def test_does_not_include_events_when_ml_insights_events_enabled_is_false():
    """
    Verifies that all ml events can be disabled by setting
    custom_insights_events.enabled.
    """

    @validate_ml_event_count(count=0)
    @background_task()
    def _test():
        import sklearn.tree

        x_train = np.array([[20, 20], [21, 21]], dtype="<U4")
        y_train = np.array([20, 21], dtype="<U4")
        x_test = np.array([[20, 21]], dtype="<U4")
        clf = getattr(sklearn.tree, "DecisionTreeClassifier")(random_state=0)

        model = clf.fit(x_train, y_train)
        labels = model.predict(x_test)

        return model

    _test()


disabled_ml_settings = {
    "machine_learning.enabled": False,
    "machine_learning.inference_events_value.enabled": True,
    "ml_insights_events.enabled": True
}


@override_application_settings(disabled_ml_settings)
@reset_core_stats_engine()
def test_does_not_include_events_when_machine_learning_enabled_is_false():
    @validate_ml_event_count(count=0)
    @background_task()
    def _test():
        import sklearn.tree

        x_train = np.array([[20, 20], [21, 21]], dtype="<U4")
        y_train = np.array([20, 21], dtype="<U4")
        x_test = np.array([[20, 21]], dtype="<U4")
        clf = getattr(sklearn.tree, "DecisionTreeClassifier")(random_state=0)

        model = clf.fit(x_train, y_train)
        labels = model.predict(x_test)

        return model

    _test()


multilabel_output_label_events = [
    (
        {"type": "InferenceData"},
        {
            "inference_id": None,
            "prediction_id": None,
            "modelName": "MultiOutputClassifier",
            "model_version": "0.0.0",
            "label.0": "1",
            "label.1": "0",
            "label.2": "1",
            "feature.0": 3.0,
            "feature.1": 5.0,
            "feature.2": 3.0,
            "feature.3": 0.0,
            "feature.4": 0.0,
            "feature.5": 2.0,
            "feature.6": 2.0,
            "feature.7": 0.0,
            "feature.8": 4.0,
            "feature.9": 3.0,
            "feature.10": 2.0,
            "feature.11": 5.0,
            "feature.12": 2.0,
            "feature.13": 3.0,
            "feature.14": 3.0,
            "feature.15": 4.0,
            "feature.16": 3.0,
            "feature.17": 1.0,
            "feature.18": 2.0,
            "feature.19": 4.0,
            "new_relic_data_schema_version": 2,
        },
    ),
]


@reset_core_stats_engine()
def test_custom_event_count_multilabel_output():
    @validate_ml_events(multilabel_output_label_events)
    @validate_ml_event_count(count=1)
    @background_task()
    def _test():
        from sklearn.datasets import make_multilabel_classification
        from sklearn.linear_model import LogisticRegression
        from sklearn.multioutput import MultiOutputClassifier

        x_train, y_train = make_multilabel_classification(n_classes=3, random_state=0)
        clf = MultiOutputClassifier(LogisticRegression()).fit(x_train, y_train)
        clf.predict([x_train[-1]])

    _test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pwd

USER = pwd.getpwuid(os.getuid()).pw_name


def postgresql_settings():
    """Return a list of dict of settings for connecting to postgresql.

    Will return the correct settings, depending on which of the environments it
    is running in. It attempts to set variables in the following order, where
    later environments override earlier ones.

        1. Local
        2. Github Actions
    """

    host = "host.docker.internal" if "GITHUB_ACTIONS" in os.environ else "localhost"
    instances = 2
    settings = [
        {
            "user": "postgres",
            "password": "postgres",
            "name": "postgres",
            "host": host,
            "port": 8080 + instance_num,
            "table_name": "postgres_table_" + str(os.getpid()),
        }
        for instance_num in range(instances)
    ]
    return settings


def mysql_settings():
    """Return a list of dict of settings for connecting to MySQL.

    Will return the correct settings, depending on which of the environments it
    is running in. It attempts to set variables in the following order, where
    later environments override earlier ones.

        1. Local
        2. Github Actions
    """

    host = "host.docker.internal" if "GITHUB_ACTIONS" in os.environ else "127.0.0.1"
    instances = 1
    settings = [
        {
            "user": "python_agent",
            "password": "python_agent",
            "name": "python_agent",
            "host": host,
            "port": 8080 + instance_num,
            "namespace": str(os.getpid()),
        }
        for instance_num in range(instances)
    ]
    return settings


def mssql_settings():
    """Return a list of dict of settings for connecting to MS SQL.

    Will return the correct settings, depending on which of the environments it
    is running in. It attempts to set variables in the following order, where
    later environments override earlier ones.

        1. Local
        2. Github Actions
    """

    host = "host.docker.internal" if "GITHUB_ACTIONS" in os.environ else "127.0.0.1"
    instances = 1
    settings = [
        {
            "user": "SA",
            "password": "python_agent#1234",
            "host": host,
            "port": 8080 + instance_num,
            "namespace": str(os.getpid()),
        }
        for instance_num in range(instances)
    ]
    return settings


def redis_settings():
    """Return a list of dict of settings for connecting to redis.

    Will return the correct settings, depending on which of the environments it
    is running in. It attempts to set variables in the following order, where
    later environments override earlier ones.

        1. Local
        2. Github Actions
    """

    host = "host.docker.internal" if "GITHUB_ACTIONS" in os.environ else "localhost"
    instances = 2
    settings = [
        {
            "host": host,
            "port": 8080 + instance_num,
        }
        for instance_num in range(instances)
    ]
    return settings


def redis_cluster_settings():
    """Return a list of dict of settings for connecting to redis cluster.

    Will return the correct settings, depending on which of the environments it
    is running in. It attempts to set variables in the following order, where
    later environments override earlier ones.

        1. Local
        2. Github Actions
    """

    host = "host.docker.internal" if "GITHUB_ACTIONS" in os.environ else "localhost"
    instances = 1
    base_port = 6379

    settings = [
        {
            "host": host,
            "port": base_port + instance_num,
        }
        for instance_num in range(instances)
    ]
    return settings


def memcached_settings():
    """Return a list of dict of settings for connecting to memcached.

    Will return the correct settings, depending on which of the environments it
    is running in. It attempts to set variables in the following order, where
    later environments override earlier ones.

        1. Local
        2. Github Actions
    """

    host = "host.docker.internal" if "GITHUB_ACTIONS" in os.environ else "127.0.0.1"
    instances = 2
    settings = [
        {
            "host": host,
            "port": 8080 + instance_num,
            "namespace": str(os.getpid()),
        }
        for instance_num in range(instances)
    ]
    return settings


def mongodb_settings():
    """Return a list of dict of settings for connecting to mongodb.

    Will return the correct settings, depending on which of the environments it
    is running in. It attempts to set variables in the following order, where
    later environments override earlier ones.

        1. Local
        2. Github Actions
    """

    host = "host.docker.internal" if "GITHUB_ACTIONS" in os.environ else "127.0.0.1"
    instances = 2
    settings = [
        {"host": host, "port": 8080 + instance_num, "collection": "mongodb_collection_" + str(os.getpid())}
        for instance_num in range(instances)
    ]
    return settings


def firestore_settings():
    """Return a list of dict of settings for connecting to firestore.

    This only includes the host and port as the collection name is defined in
    the firestore conftest file.
    Will return the correct settings, depending on which of the environments it
    is running in. It attempts to set variables in the following order, where
    later environments override earlier ones.

        1. Local
        2. Github Actions
    """

    host = "host.docker.internal" if "GITHUB_ACTIONS" in os.environ else "127.0.0.1"
    instances = 2
    settings = [
        {"host": host, "port": 8080 + instance_num}
        for instance_num in range(instances)
    ]
    return settings


def elasticsearch_settings():
    """Return a list of dict of settings for connecting to elasticsearch.

    Will return the correct settings, depending on which of the environments it
    is running in. It attempts to set variables in the following order, where
    later environments override earlier ones.

        1. Local
        2. Github Actions
    """

    host = "host.docker.internal" if "GITHUB_ACTIONS" in os.environ else "localhost"
    instances = 2
    settings = [
        {
            "host": host,
            "port": str(8080 + instance_num),
            "namespace": str(os.getpid()),
        }
        for instance_num in range(instances)
    ]
    return settings


def solr_settings():
    """Return a list of dict of settings for connecting to solr.

    Will return the correct settings, depending on which of the environments it
    is running in. It attempts to set variables in the following order, where
    later environments override earlier ones.

        1. Local
        2. Github Actions
    """

    host = "host.docker.internal" if "GITHUB_ACTIONS" in os.environ else "127.0.0.1"
    instances = 2
    settings = [
        {
            "host": host,
            "port": 8080 + instance_num,
            "namespace": str(os.getpid()),
        }
        for instance_num in range(instances)
    ]
    return settings


def rabbitmq_settings():
    """Return a list of dict of settings for connecting to rabbitmq.

    Will return the correct settings, depending on which of the environments it
    is running in. It attempts to set variables in the following order, where
    later environments override earlier ones.

        1. Local
        2. Github Actions
    """

    host = "host.docker.internal" if "GITHUB_ACTIONS" in os.environ else "localhost"
    instances = 1
    settings = [
        {
            "host": host,
            "port": 5672 + instance_num,
        }
        for instance_num in range(instances)
    ]
    return settings


def kafka_settings():
    """Return a list of dict of settings for connecting to kafka.

    Will return the correct settings, depending on which of the environments it
    is running in. It attempts to set variables in the following order, where
    later environments override earlier ones.

        1. Local
        2. Github Actions
    """

    host = "host.docker.internal" if "GITHUB_ACTIONS" in os.environ else "127.0.0.1"
    base_port = 8082 if "GITHUB_ACTIONS" in os.environ else 8080
    instances = 2
    settings = [
        {
            "host": host,
            "port": base_port + instance_num,
        }
        for instance_num in range(instances)
    ]
    return settings


def gearman_settings():
    """Return a list of dict of settings for connecting to kafka.

    Will return the correct settings, depending on which of the environments it
    is running in. It attempts to set variables in the following order, where
    later environments override earlier ones.

        1. Local
        2. Github Actions
    """

    host = "host.docker.internal" if "GITHUB_ACTIONS" in os.environ else "localhost"
    instances = 1
    settings = [
        {
            "host": host,
            "port": 8080 + instance_num,
        }
        for instance_num in range(instances)
    ]
    return settings



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

import pytest
from testing_support.fixtures import override_generic_settings
from testing_support.validators.validate_serverless_data import validate_serverless_data
from testing_support.validators.validate_serverless_metadata import (
    validate_serverless_metadata,
)
from testing_support.validators.validate_serverless_payload import (
    validate_serverless_payload,
)

from newrelic.api.application import application_instance
from newrelic.api.background_task import background_task
from newrelic.api.external_trace import ExternalTrace
from newrelic.api.lambda_handler import lambda_handler
from newrelic.api.transaction import current_transaction
from newrelic.core.config import global_settings


@pytest.fixture(scope="function")
def serverless_application(request):
    settings = global_settings()
    orig = settings.serverless_mode.enabled
    settings.serverless_mode.enabled = True

    application_name = "Python Agent Test (test_serverless_mode:%s)" % (request.node.name)
    application = application_instance(application_name)
    application.activate()

    yield application

    settings.serverless_mode.enabled = orig


def test_serverless_payload(capsys, serverless_application):
    @override_generic_settings(
        serverless_application.settings,
        {
            "distributed_tracing.enabled": True,
        },
    )
    @validate_serverless_data(
        expected_methods=("metric_data", "analytic_event_data"),
        forgone_methods=("preconnect", "connect", "get_agent_commands"),
    )
    @validate_serverless_payload()
    @background_task(application=serverless_application, name="test_serverless_payload")
    def _test():
        transaction = current_transaction()
        assert transaction.settings.serverless_mode.enabled

    _test()

    out, err = capsys.readouterr()

    # Validate that something is printed to stdout
    assert out

    # Verify that the payload is loadable JSON
    payload = json.loads(out)


def test_no_cat_headers(serverless_application):
    @background_task(application=serverless_application, name="test_cat_headers")
    def _test_cat_headers():
        transaction = current_transaction()

        payload = ExternalTrace.generate_request_headers(transaction)
        assert not payload

        trace = ExternalTrace("testlib", "http://example.com")
        response_headers = [("X-NewRelic-App-Data", "Cookies")]
        with trace:
            trace.process_response_headers(response_headers)

        assert transaction.settings.cross_application_tracer.enabled is False

    _test_cat_headers()


@pytest.mark.parametrize("trusted_account_key", ("1", None), ids=("tk_set", "tk_unset"))
def test_outbound_dt_payload_generation(serverless_application, trusted_account_key):
    @override_generic_settings(
        serverless_application.settings,
        {
            "distributed_tracing.enabled": True,
            "account_id": "1",
            "trusted_account_key": trusted_account_key,
            "primary_application_id": "1",
        },
    )
    @background_task(application=serverless_application, name="test_outbound_dt_payload_generation")
    def _test_outbound_dt_payload_generation():
        transaction = current_transaction()
        payload = ExternalTrace.generate_request_headers(transaction)
        assert payload
        # Ensure trusted account key or account ID present as vendor
        assert dict(payload)["tracestate"].startswith("1@nr=")

    _test_outbound_dt_payload_generation()


@pytest.mark.parametrize("trusted_account_key", ("1", None), ids=("tk_set", "tk_unset"))
def test_inbound_dt_payload_acceptance(serverless_application, trusted_account_key):
    @override_generic_settings(
        serverless_application.settings,
        {
            "distributed_tracing.enabled": True,
            "account_id": "1",
            "trusted_account_key": trusted_account_key,
            "primary_application_id": "1",
        },
    )
    @background_task(application=serverless_application, name="test_inbound_dt_payload_acceptance")
    def _test_inbound_dt_payload_acceptance():
        transaction = current_transaction()

        payload = {
            "v": [0, 1],
            "d": {
                "ty": "Mobile",
                "ac": "1",
                "tk": "1",
                "ap": "2827902",
                "pa": "5e5733a911cfbc73",
                "id": "7d3efb1b173fecfa",
                "tr": "d6b4ba0c3a712ca",
                "ti": 1518469636035,
                "tx": "8703ff3d88eefe9d",
            },
        }

        result = transaction.accept_distributed_trace_payload(payload)
        assert result

    _test_inbound_dt_payload_acceptance()


# The lambda_hander has been deprecated for 3+ years
@pytest.mark.skip(reason="The lambda_handler has been deprecated")
@pytest.mark.parametrize("arn_set", (True, False))
def test_payload_metadata_arn(serverless_application, arn_set):
    # If the session object gathers the arn from the settings object before the
    # lambda handler records it there, then this test will fail.

    settings = global_settings()
    original_metadata = settings.aws_lambda_metadata.copy()

    arn = None
    if arn_set:
        arn = "arrrrrrrrrrRrrrrrrrn"

    settings.aws_lambda_metadata.update({"arn": arn, "function_version": "$LATEST"})

    class Context(object):
        invoked_function_arn = arn

    @validate_serverless_metadata(exact_metadata={"arn": arn})
    @lambda_handler(application=serverless_application)
    def handler(event, context):
        assert settings.aws_lambda_metadata["arn"] == arn
        return {}

    try:
        handler({}, Context)
    finally:
        settings.aws_lambda_metadata = original_metadata



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    import http.client as httplib
except ImportError:
    import httplib


from newrelic.api.external_trace import ExternalTrace
from newrelic.api.transaction import current_transaction
from newrelic.common.object_wrapper import transient_function_wrapper
from newrelic.common.encoding_utils import json_encode, obfuscate


def create_incoming_headers(transaction):
    settings = transaction.settings
    encoding_key = settings.encoding_key

    headers = []

    cross_process_id = '1#2'
    path = 'test'
    queue_time = 1.0
    duration = 2.0
    read_length = 1024
    guid = '0123456789012345'
    record_tt = False

    payload = (cross_process_id, path, queue_time, duration, read_length,
            guid, record_tt)
    app_data = json_encode(payload)

    value = obfuscate(app_data, encoding_key)

    assert isinstance(value, type(''))

    headers.append(('X-NewRelic-App-Data', value))

    return headers


def validate_synthetics_external_trace_header(
        synthetics_header,
        synthetics_info_header,
    ):
    @transient_function_wrapper('newrelic.core.stats_engine',
            'StatsEngine.record_transaction')
    def _validate_synthetics_external_trace_header(wrapped, instance,
            args, kwargs):
        def _bind_params(transaction, *args, **kwargs):
            return transaction

        transaction = _bind_params(*args, **kwargs)

        try:
            result = wrapped(*args, **kwargs)
        except:
            raise
        else:
            # XXX This validation routine is technically
            # broken as the argument to record_transaction()
            # is not actually an instance of the Transaction
            # object. Instead it is a TransactionNode object.
            # The static method generate_request_headers() is
            # expecting a Transaction object and not
            # TransactionNode. The latter provides attributes
            # which are not updatable by the static method
            # generate_request_headers(), which it wants to
            # update, so would fail. For now what we do is use
            # a little proxy wrapper so that updates do not
            # fail. The use of this wrapper needs to be
            # reviewed and a better way of achieving what is
            # required found.

            class _Transaction(object):
                def __init__(self, wrapped):
                    self.__wrapped__ = wrapped

                def __getattr__(self, name):
                    return getattr(self.__wrapped__, name, lambda *args, **kwargs: None)

            external_headers = ExternalTrace.generate_request_headers(
                    _Transaction(transaction))
            external_headers = {header[0]: header[1] for header in external_headers}

            if synthetics_header:
                assert synthetics_header == external_headers["X-NewRelic-Synthetics"], (
                        'synthetics_header=%r, external_headers=%r' % (
                        synthetics_header, external_headers))
            else:
                assert "X-NewRelic-Synthetics" not in external_headers

            if synthetics_info_header:
                assert synthetics_info_header == external_headers["X-NewRelic-Synthetics-Info"], (
                        'synthetics_info_header=%r, external_headers=%r' % (
                        synthetics_info_header, external_headers))
            else:
                assert "X-NewRelic-Synthetics-Info" not in external_headers


        return result

    return _validate_synthetics_external_trace_header


@transient_function_wrapper(httplib.__name__, 'HTTPConnection.putheader')
def cache_outgoing_headers(wrapped, instance, args, kwargs):
    def _bind_params(header, *values):
        return header, values

    transaction = current_transaction()

    if transaction is None:
        return wrapped(*args, **kwargs)

    header, values = _bind_params(*args, **kwargs)

    try:
        cache = transaction._test_request_headers
    except AttributeError:
        cache = transaction._test_request_headers = {}

    try:
        cache[header].extend(values)
    except KeyError:
        cache[header] = list(values)

    return wrapped(*args, **kwargs)


@transient_function_wrapper(httplib.__name__, 'HTTPResponse.getheaders')
def insert_incoming_headers(wrapped, instance, args, kwargs):
    transaction = current_transaction()

    if transaction is None:
        return wrapped(*args, **kwargs)

    headers = list(wrapped(*args, **kwargs))

    headers.extend(create_incoming_headers(transaction))

    return headers


# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import logging
import sys

_builtin_plugins = [
    "debug_console",
    "generate_config",
    "license_key",
    "local_config",
    "network_config",
    "record_deploy",
    "run_program",
    "run_python",
    "server_config",
    "validate_config",
]

_commands = {}


def command(name, options="", description="", hidden=False, log_intercept=True, deprecated=False):
    def wrapper(callback):
        callback.name = name
        callback.options = options
        callback.description = description
        callback.hidden = hidden
        callback.log_intercept = log_intercept
        callback.deprecated = deprecated
        _commands[name] = callback
        return callback

    return wrapper


def usage(name):
    details = _commands[name]
    if details.deprecated:
        print("[WARNING] This command is deprecated and will be removed")
    print("Usage: newrelic-admin %s %s" % (name, details.options))


@command("help", "[command]", hidden=True)
def help(args):
    if not args:
        print("Usage: newrelic-admin command [options]")
        print()
        print("Type 'newrelic-admin help <command>'", end="")
        print("for help on a specific command.")
        print()
        print("Available commands are:")

        commands = sorted(_commands.keys())
        for name in commands:
            details = _commands[name]
            if not details.hidden:
                print(" ", name)

    else:
        name = args[0]

        if name not in _commands:
            print("Unknown command '%s'." % name, end=" ")
            print("Type 'newrelic-admin help' for usage.")

        else:
            details = _commands[name]

            print("Usage: newrelic-admin %s %s" % (name, details.options))
            if details.description:
                print()
                description = details.description
                if details.deprecated:
                    description = "[DEPRECATED] " + description
                print(description)


def setup_log_intercept():
    # Send any errors or warnings to standard output as well so more
    # obvious as use may have trouble finding the relevant messages in
    # the agent log as it will be full of debug output as well.

    class FilteredStreamHandler(logging.StreamHandler):
        def emit(self, record):
            if len(logging.root.handlers) != 0:
                return

            if record.name.startswith("newrelic.packages"):
                return

            if record.levelno < logging.WARNING:
                return

            return logging.StreamHandler.emit(self, record)

    _stdout_logger = logging.getLogger("newrelic")
    _stdout_handler = FilteredStreamHandler(sys.stdout)
    _stdout_format = "%(levelname)s - %(message)s\n"
    _stdout_formatter = logging.Formatter(_stdout_format)
    _stdout_handler.setFormatter(_stdout_formatter)
    _stdout_logger.addHandler(_stdout_handler)


def load_internal_plugins():
    for name in _builtin_plugins:
        module_name = "%s.%s" % (__name__, name)
        __import__(module_name)


def load_external_plugins():
    try:
        # Preferred after Python 3.10
        if sys.version_info >= (3, 10):
            from importlib.metadata import entry_points
        # Introduced in Python 3.8
        elif sys.version_info >= (3, 8) and sys.version_info <= (3, 9):
            from importlib_metadata import entry_points
        # Removed in Python 3.12
        else:
            from pkg_resources import iter_entry_points as entry_points
    except ImportError:
        return

    group = "newrelic.admin"

    for entrypoint in entry_points(group=group):
        __import__(entrypoint.module_name)


def main():
    try:
        if len(sys.argv) > 1:
            command = sys.argv[1]
        else:
            command = "help"

        callback = _commands[command]

    except Exception:
        print("Unknown command '%s'." % command, end="")
        print("Type 'newrelic-admin help' for usage.")
        sys.exit(1)

    if callback.log_intercept:
        setup_log_intercept()

    callback(sys.argv[2:])


load_internal_plugins()
load_external_plugins()

if __name__ == "__main__":
    main()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.api.asgi_application import ASGIApplicationWrapper
from newrelic.api.wsgi_application import WSGIApplicationWrapper
from newrelic.common.object_wrapper import wrap_function_wrapper
from newrelic.common.package_version_utils import get_package_version


def bind_worker_serve(app, *args, **kwargs):
    return app, args, kwargs


async def wrap_worker_serve(wrapped, instance, args, kwargs):
    import hypercorn

    dispatcher_details = ("Hypercorn", get_package_version("hypercorn"))
    wrapper_module = getattr(hypercorn, "app_wrappers", None)
    asgi_wrapper_class = getattr(wrapper_module, "ASGIWrapper", None)
    wsgi_wrapper_class = getattr(wrapper_module, "WSGIWrapper", None)

    app, args, kwargs = bind_worker_serve(*args, **kwargs)

    # Hypercorn 0.14.1 introduced wrappers for ASGI and WSGI apps that need to be above our instrumentation.
    if asgi_wrapper_class is not None and isinstance(app, asgi_wrapper_class):
        app.app = ASGIApplicationWrapper(app.app, dispatcher=dispatcher_details)
    elif wsgi_wrapper_class is not None and isinstance(app, wsgi_wrapper_class):
        app.app = WSGIApplicationWrapper(app.app, dispatcher=dispatcher_details)
    else:
        app = ASGIApplicationWrapper(app, dispatcher=dispatcher_details)

    app._nr_wrapped = True

    return await wrapped(app, *args, **kwargs)


def bind_is_asgi(app):
    return app


def wrap_is_asgi(wrapped, instance, args, kwargs):
    # Wrapper is identical and reused for the functions is_asgi and _is_asgi_2.
    app = bind_is_asgi(*args, **kwargs)

    # Unwrap apps wrapped by our instrumentation.
    # ASGI 2/3 detection for hypercorn is unable to process
    # our wrappers and will return incorrect results. This
    # should be sufficient to allow hypercorn to run detection
    # on an application that was not wrapped by this instrumentation.
    while getattr(app, "_nr_wrapped", False):
        app = app.__wrapped__

    return wrapped(app)


def instrument_hypercorn_asyncio_run(module):
    if hasattr(module, "worker_serve"):
        wrap_function_wrapper(module, "worker_serve", wrap_worker_serve)


def instrument_hypercorn_trio_run(module):
    if hasattr(module, "worker_serve"):
        wrap_function_wrapper(module, "worker_serve", wrap_worker_serve)


def instrument_hypercorn_utils(module):
    if hasattr(module, "_is_asgi_2"):
        wrap_function_wrapper(module, "_is_asgi_2", wrap_is_asgi)

    if hasattr(module, "is_asgi"):
        wrap_function_wrapper(module, "is_asgi", wrap_is_asgi)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from newrelic.api.background_task import background_task

from testing_support.validators.validate_sql_obfuscation import validate_sql_obfuscation

@pytest.fixture()
def sqlite3_cursor():
    import sqlite3

    with sqlite3.connect(':memory:') as connection:
        cursor = connection.cursor()
        cursor.execute('drop table if exists a')
        cursor.execute('create table a (b, c)')
        cursor.executemany('insert into a values (?, ?)',
                [(1, 1), (2, 2), (3, 3)])

        yield cursor


_quoting_style_tests = [
    ('SELECT * FROM a WHERE b="c"', 'SELECT * FROM a WHERE b=?'),
    ("SELECT * FROM a WHERE b='c'", 'SELECT * FROM a WHERE b=?'),
]


@pytest.mark.parametrize('sql,obfuscated', _quoting_style_tests)
def test_quoting_styles(sqlite3_cursor, sql, obfuscated):

    @validate_sql_obfuscation([obfuscated])
    @background_task()
    def test():
        sqlite3_cursor.execute(sql)

    test()


_parameter_tests = [
    ('INSERT INTO a VALUES (:1, :2)', 'INSERT INTO a VALUES (:1, :2)'),
    ('INSERT INTO a VALUES (?, ?)', 'INSERT INTO a VALUES (?, ?)'),
]


@pytest.mark.parametrize('sql,obfuscated', _parameter_tests)
def test_parameters(sqlite3_cursor, sql, obfuscated):

    @validate_sql_obfuscation([obfuscated])
    @background_task()
    def test():
        sqlite3_cursor.executemany(sql,
                [('hello', 'world'), ('love', 'python')])

    test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import gc
import pika

from newrelic.api.background_task import background_task
from testing_support.db_settings import rabbitmq_settings


DB_SETTINGS = rabbitmq_settings()[0]


@background_task()
def test_memory_leak():
    params = pika.ConnectionParameters(
            DB_SETTINGS['host'], DB_SETTINGS['port'])

    # create 2 unreferenced blocking channels
    with pika.BlockingConnection(params) as connection:
        for _ in range(2):
            connection.channel().basic_publish(
                    exchange='', routing_key='memory_leak_test', body='test')

    # garbage collect until everything is reachable
    while gc.collect():
        pass

    # the number of channel objects stored should be 0
    from pika.adapters.blocking_connection import BlockingChannel
    channel_objects_stored = sum(1 for o in gc.get_objects()
            if isinstance(o, BlockingChannel))

    assert channel_objects_stored == 0



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.background_task import background_task
from newrelic.common.package_version_utils import get_package_version
from newrelic.packages import six

SKLEARN_VERSION = tuple(map(int, get_package_version("sklearn").split(".")))


@pytest.mark.parametrize(
    "ensemble_model_name",
    [
        "AdaBoostClassifier",
        "AdaBoostRegressor",
        "BaggingClassifier",
        "BaggingRegressor",
        "ExtraTreesClassifier",
        "ExtraTreesRegressor",
        "GradientBoostingClassifier",
        "GradientBoostingRegressor",
        "IsolationForest",
        "RandomForestClassifier",
        "RandomForestRegressor",
        "RandomTreesEmbedding",
        "VotingClassifier",
    ],
)
def test_below_v1_0_model_methods_wrapped_in_function_trace(ensemble_model_name, run_ensemble_model):
    expected_scoped_metrics = {
        "AdaBoostClassifier": [
            ("Function/MLModel/Sklearn/Named/AdaBoostClassifier.fit", 1),
            ("Function/MLModel/Sklearn/Named/AdaBoostClassifier.predict", 2),
            ("Function/MLModel/Sklearn/Named/AdaBoostClassifier.predict_log_proba", 1),
            ("Function/MLModel/Sklearn/Named/AdaBoostClassifier.predict_proba", 2),
            ("Function/MLModel/Sklearn/Named/AdaBoostClassifier.score", 1),
        ],
        "AdaBoostRegressor": [
            ("Function/MLModel/Sklearn/Named/AdaBoostRegressor.fit", 1),
            ("Function/MLModel/Sklearn/Named/AdaBoostRegressor.predict", 2),
            ("Function/MLModel/Sklearn/Named/AdaBoostRegressor.score", 1),
        ],
        "BaggingClassifier": [
            ("Function/MLModel/Sklearn/Named/BaggingClassifier.fit", 1),
            ("Function/MLModel/Sklearn/Named/BaggingClassifier.predict", 2),
            ("Function/MLModel/Sklearn/Named/BaggingClassifier.score", 1),
            ("Function/MLModel/Sklearn/Named/BaggingClassifier.predict_log_proba", 1),
            ("Function/MLModel/Sklearn/Named/BaggingClassifier.predict_proba", 3),
        ],
        "BaggingRegressor": [
            ("Function/MLModel/Sklearn/Named/BaggingRegressor.fit", 1),
            ("Function/MLModel/Sklearn/Named/BaggingRegressor.predict", 2),
            ("Function/MLModel/Sklearn/Named/BaggingRegressor.score", 1),
        ],
        "ExtraTreesClassifier": [
            ("Function/MLModel/Sklearn/Named/ExtraTreesClassifier.fit", 1),
            ("Function/MLModel/Sklearn/Named/ExtraTreesClassifier.predict", 2),
            ("Function/MLModel/Sklearn/Named/ExtraTreesClassifier.score", 1),
            ("Function/MLModel/Sklearn/Named/ExtraTreesClassifier.predict_log_proba", 1),
            ("Function/MLModel/Sklearn/Named/ExtraTreesClassifier.predict_proba", 4),
        ],
        "ExtraTreesRegressor": [
            ("Function/MLModel/Sklearn/Named/ExtraTreesRegressor.fit", 1),
            ("Function/MLModel/Sklearn/Named/ExtraTreesRegressor.predict", 2),
            ("Function/MLModel/Sklearn/Named/ExtraTreesRegressor.score", 1),
        ],
        "GradientBoostingClassifier": [
            ("Function/MLModel/Sklearn/Named/GradientBoostingClassifier.fit", 1),
            ("Function/MLModel/Sklearn/Named/GradientBoostingClassifier.predict", 2),
            ("Function/MLModel/Sklearn/Named/GradientBoostingClassifier.score", 1),
            ("Function/MLModel/Sklearn/Named/GradientBoostingClassifier.predict_log_proba", 1),
            ("Function/MLModel/Sklearn/Named/GradientBoostingClassifier.predict_proba", 2),
        ],
        "GradientBoostingRegressor": [
            ("Function/MLModel/Sklearn/Named/GradientBoostingRegressor.fit", 1),
            ("Function/MLModel/Sklearn/Named/GradientBoostingRegressor.predict", 2),
            ("Function/MLModel/Sklearn/Named/GradientBoostingRegressor.score", 1),
        ],
        "IsolationForest": [
            ("Function/MLModel/Sklearn/Named/IsolationForest.fit", 1),
            ("Function/MLModel/Sklearn/Named/IsolationForest.predict", 1),
        ],
        "RandomForestClassifier": [
            ("Function/MLModel/Sklearn/Named/RandomForestClassifier.fit", 1),
            ("Function/MLModel/Sklearn/Named/RandomForestClassifier.predict", 2),
            ("Function/MLModel/Sklearn/Named/RandomForestClassifier.score", 1),
            ("Function/MLModel/Sklearn/Named/RandomForestClassifier.predict_log_proba", 1),
            ("Function/MLModel/Sklearn/Named/RandomForestClassifier.predict_proba", 4),
        ],
        "RandomForestRegressor": [
            ("Function/MLModel/Sklearn/Named/RandomForestRegressor.fit", 1),
            ("Function/MLModel/Sklearn/Named/RandomForestRegressor.predict", 2),
            ("Function/MLModel/Sklearn/Named/RandomForestRegressor.score", 1),
        ],
        "RandomTreesEmbedding": [
            ("Function/MLModel/Sklearn/Named/RandomTreesEmbedding.fit", 1),
            ("Function/MLModel/Sklearn/Named/RandomTreesEmbedding.transform", 1),
        ],
        "VotingClassifier": [
            ("Function/MLModel/Sklearn/Named/VotingClassifier.fit", 1),
            ("Function/MLModel/Sklearn/Named/VotingClassifier.predict", 2),
            ("Function/MLModel/Sklearn/Named/VotingClassifier.score", 1),
            ("Function/MLModel/Sklearn/Named/VotingClassifier.transform", 1),
            ("Function/MLModel/Sklearn/Named/VotingClassifier.predict_proba", 3),
        ],
    }

    expected_transaction_name = (
        "test_ensemble_models:test_below_v1_0_model_methods_wrapped_in_function_trace.<locals>._test"
        if six.PY3
        else "test_ensemble_models:_test"
    )

    @validate_transaction_metrics(
        expected_transaction_name,
        scoped_metrics=expected_scoped_metrics[ensemble_model_name],
        rollup_metrics=expected_scoped_metrics[ensemble_model_name],
        background_task=True,
    )
    @background_task()
    def _test():
        run_ensemble_model(ensemble_model_name)

    _test()


@pytest.mark.skipif(SKLEARN_VERSION < (1, 0, 0) or SKLEARN_VERSION >= (1, 1, 0), reason="Requires 1.0 <= sklearn < 1.1")
@pytest.mark.parametrize(
    "ensemble_model_name",
    [
        "HistGradientBoostingClassifier",
        "HistGradientBoostingRegressor",
        "StackingClassifier",
        "StackingRegressor",
        "VotingRegressor",
    ],
)
def test_between_v1_0_and_v1_1_model_methods_wrapped_in_function_trace(ensemble_model_name, run_ensemble_model):
    expected_scoped_metrics = {
        "HistGradientBoostingClassifier": [
            ("Function/MLModel/Sklearn/Named/HistGradientBoostingClassifier.fit", 1),
            ("Function/MLModel/Sklearn/Named/HistGradientBoostingClassifier.predict", 2),
            ("Function/MLModel/Sklearn/Named/HistGradientBoostingClassifier.score", 1),
            ("Function/MLModel/Sklearn/Named/HistGradientBoostingClassifier.predict_proba", 3),
        ],
        "HistGradientBoostingRegressor": [
            ("Function/MLModel/Sklearn/Named/HistGradientBoostingRegressor.fit", 1),
            ("Function/MLModel/Sklearn/Named/HistGradientBoostingRegressor.predict", 2),
            ("Function/MLModel/Sklearn/Named/HistGradientBoostingRegressor.score", 1),
        ],
        "StackingClassifier": [
            ("Function/MLModel/Sklearn/Named/StackingClassifier.fit", 1),
        ],
        "StackingRegressor": [
            ("Function/MLModel/Sklearn/Named/StackingRegressor.fit", 1),
        ],
        "VotingRegressor": [
            ("Function/MLModel/Sklearn/Named/VotingRegressor.fit", 1),
            ("Function/MLModel/Sklearn/Named/VotingRegressor.predict", 2),
            ("Function/MLModel/Sklearn/Named/VotingRegressor.score", 1),
            ("Function/MLModel/Sklearn/Named/VotingRegressor.transform", 1),
        ],
    }
    expected_transaction_name = (
        "test_ensemble_models:test_between_v1_0_and_v1_1_model_methods_wrapped_in_function_trace.<locals>._test"
        if six.PY3
        else "test_ensemble_models:_test"
    )

    @validate_transaction_metrics(
        expected_transaction_name,
        scoped_metrics=expected_scoped_metrics[ensemble_model_name],
        rollup_metrics=expected_scoped_metrics[ensemble_model_name],
        background_task=True,
    )
    @background_task()
    def _test():
        run_ensemble_model(ensemble_model_name)

    _test()



@pytest.mark.skipif(SKLEARN_VERSION < (1, 1, 0), reason="Requires sklearn >= 1.1")
@pytest.mark.parametrize(
    "ensemble_model_name",
    [
        "HistGradientBoostingClassifier",
        "HistGradientBoostingRegressor",
        "StackingClassifier",
        "StackingRegressor",
        "VotingRegressor",
    ],
)
def test_above_v1_1_model_methods_wrapped_in_function_trace(ensemble_model_name, run_ensemble_model):
    expected_scoped_metrics = {
        "StackingClassifier": [
            ("Function/MLModel/Sklearn/Named/StackingClassifier.fit", 1),
            ("Function/MLModel/Sklearn/Named/StackingClassifier.predict", 2),
            ("Function/MLModel/Sklearn/Named/StackingClassifier.score", 1),
            ("Function/MLModel/Sklearn/Named/StackingClassifier.predict_proba", 1),
            ("Function/MLModel/Sklearn/Named/StackingClassifier.transform", 4),
        ],
        "StackingRegressor": [
            ("Function/MLModel/Sklearn/Named/StackingRegressor.fit", 1),
            ("Function/MLModel/Sklearn/Named/StackingRegressor.predict", 2),
            ("Function/MLModel/Sklearn/Named/StackingRegressor.score", 1),
        ],
        "VotingRegressor": [
            ("Function/MLModel/Sklearn/Named/VotingRegressor.fit", 1),
            ("Function/MLModel/Sklearn/Named/VotingRegressor.predict", 2),
            ("Function/MLModel/Sklearn/Named/VotingRegressor.score", 1),
            ("Function/MLModel/Sklearn/Named/VotingRegressor.transform", 1),
        ],
        "HistGradientBoostingClassifier": [
            ("Function/MLModel/Sklearn/Named/HistGradientBoostingClassifier.fit", 1),
            ("Function/MLModel/Sklearn/Named/HistGradientBoostingClassifier.predict", 2),
            ("Function/MLModel/Sklearn/Named/HistGradientBoostingClassifier.score", 1),
            ("Function/MLModel/Sklearn/Named/HistGradientBoostingClassifier.predict_proba", 3),
        ],
        "HistGradientBoostingRegressor": [
            ("Function/MLModel/Sklearn/Named/HistGradientBoostingRegressor.fit", 1),
            ("Function/MLModel/Sklearn/Named/HistGradientBoostingRegressor.predict", 2),
            ("Function/MLModel/Sklearn/Named/HistGradientBoostingRegressor.score", 1),
        ],
    }
    expected_transaction_name = (
        "test_ensemble_models:test_above_v1_1_model_methods_wrapped_in_function_trace.<locals>._test"
        if six.PY3
        else "test_ensemble_models:_test"
    )

    @validate_transaction_metrics(
        expected_transaction_name,
        scoped_metrics=expected_scoped_metrics[ensemble_model_name],
        rollup_metrics=expected_scoped_metrics[ensemble_model_name],
        background_task=True,
    )
    @background_task()
    def _test():
        run_ensemble_model(ensemble_model_name)

    _test()


@pytest.fixture
def run_ensemble_model():
    def _run(ensemble_model_name):
        import sklearn.ensemble
        from sklearn.datasets import load_iris
        from sklearn.model_selection import train_test_split

        X, y = load_iris(return_X_y=True)
        x_train, x_test, y_train, y_test = train_test_split(X, y, stratify=y, random_state=0)

        kwargs = {"random_state": 0}
        if ensemble_model_name == "StackingClassifier":
            kwargs = {"estimators": [("rf", RandomForestClassifier())], "final_estimator": RandomForestClassifier()}
        elif ensemble_model_name == "VotingClassifier":
            kwargs = {
                "estimators": [("rf", RandomForestClassifier())],
                "voting": "soft",
            }
        elif ensemble_model_name == "VotingRegressor":
            x_train = x_test = [[1, 1]]
            y_train = y_test = [0]
            kwargs = {"estimators": [("rf", RandomForestRegressor())]}
        elif ensemble_model_name == "StackingRegressor":
            kwargs = {"estimators": [("rf", RandomForestRegressor())]}
        clf = getattr(sklearn.ensemble, ensemble_model_name)(**kwargs)

        model = clf.fit(x_train, y_train)
        if hasattr(model, "predict"):
            model.predict(x_test)
        if hasattr(model, "score"):
            model.score(x_test, y_test)
        if hasattr(model, "predict_log_proba"):
            model.predict_log_proba(x_test)
        if hasattr(model, "predict_proba"):
            model.predict_proba(x_test)
        if hasattr(model, "transform"):
            model.transform(x_test)

        return model

    return _run



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging

import pytest

from newrelic.agent import get_linking_metadata
from newrelic.api.background_task import background_task
from newrelic.api.function_trace import FunctionTrace
from newrelic.api.log import NewRelicContextFormatter
from newrelic.packages import six

if six.PY2:
    from io import BytesIO as Buffer
else:
    from io import StringIO as Buffer

_logger = logging.getLogger(__name__)


@pytest.fixture
def log_buffer(caplog):
    buf = Buffer()

    _formatter = NewRelicContextFormatter("", datefmt="ISO8601")
    _handler = logging.StreamHandler(buf)
    _handler.setFormatter(_formatter)

    _logger.addHandler(_handler)
    caplog.set_level(logging.INFO, logger=__name__)

    yield buf

    _logger.removeHandler(_handler)


class NonPrintableObject(object):
    def __str__(self):
        raise RuntimeError("Unable to print object.")

    def __repr__(self):
        raise RuntimeError("Unable to print object.")


def test_newrelic_logger_no_error(log_buffer):
    extra = {
        "string": "foo",
        "integer": 1,
        "float": 1.23,
        "null": None,
        "array": [1, 2, 3],
        "bool": True,
        "non_serializable": {"set"},
        "non_printable": NonPrintableObject(),
        "object": {
            "first": "bar",
            "second": "baz",
        },
    }
    _logger.info(u"Hello %s", u"World", extra=extra)

    log_buffer.seek(0)
    message = json.load(log_buffer)

    timestamp = message.pop("timestamp")
    thread_id = message.pop("thread.id")
    process_id = message.pop("process.id")
    filename = message.pop("file.name")
    line_number = message.pop("line.number")

    assert isinstance(timestamp, int)
    assert isinstance(thread_id, int)
    assert isinstance(process_id, int)
    assert filename.endswith("/test_logs_in_context.py")
    assert isinstance(line_number, int)

    expected = {
        u"entity.name": u"Python Agent Test (agent_features)",
        u"entity.type": u"SERVICE",
        u"message": u"Hello World",
        u"log.level": u"INFO",
        u"logger.name": u"test_logs_in_context",
        u"thread.name": u"MainThread",
        u"process.name": u"MainProcess",
        u"extra.string": u"foo",
        u"extra.integer": 1,
        u"extra.float": 1.23,
        u"extra.null": None,
        u"extra.array": [1, 2, 3],
        u"extra.bool": True,
        u"extra.non_serializable": u"set(['set'])" if six.PY2 else u"{'set'}",
        u"extra.non_printable": u"<unprintable NonPrintableObject object>",
        u"extra.object": {
            u"first": u"bar",
            u"second": u"baz",
        },
    }
    expected_extra_txn_keys = (
        "entity.guid",
        "hostname",
    )

    for k, v in expected.items():
        assert message.pop(k) == v

    assert set(message.keys()) == set(expected_extra_txn_keys)



class ExceptionForTest(ValueError):
    pass


@background_task()
def test_newrelic_logger_error_inside_transaction(log_buffer):
    try:
        raise ExceptionForTest
    except ExceptionForTest:
        _logger.exception(u"oops")

    log_buffer.seek(0)
    message = json.load(log_buffer)

    timestamp = message.pop("timestamp")
    thread_id = message.pop("thread.id")
    process_id = message.pop("process.id")
    filename = message.pop("file.name")
    line_number = message.pop("line.number")

    assert isinstance(timestamp, int)
    assert isinstance(thread_id, int)
    assert isinstance(process_id, int)
    assert filename.endswith("/test_logs_in_context.py")
    assert isinstance(line_number, int)

    expected = {
        u"entity.name": u"Python Agent Test (agent_features)",
        u"entity.type": u"SERVICE",
        u"message": u"oops",
        u"log.level": u"ERROR",
        u"logger.name": u"test_logs_in_context",
        u"thread.name": u"MainThread",
        u"process.name": u"MainProcess",
        u"error.class": u"test_logs_in_context:ExceptionForTest",
        u"error.message": u"",
        u"error.expected": False,
    }
    expected_extra_txn_keys = (
        "trace.id",
        "span.id",
        "entity.guid",
        "hostname",
    )

    for k, v in expected.items():
        assert message.pop(k) == v

    assert set(message.keys()) == set(expected_extra_txn_keys)


def test_newrelic_logger_error_outside_transaction(log_buffer):
    try:
        raise ExceptionForTest
    except ExceptionForTest:
        _logger.exception(u"oops")

    log_buffer.seek(0)
    message = json.load(log_buffer)

    timestamp = message.pop("timestamp")
    thread_id = message.pop("thread.id")
    process_id = message.pop("process.id")
    filename = message.pop("file.name")
    line_number = message.pop("line.number")

    assert isinstance(timestamp, int)
    assert isinstance(thread_id, int)
    assert isinstance(process_id, int)
    assert filename.endswith("/test_logs_in_context.py")
    assert isinstance(line_number, int)

    expected = {
        u"entity.name": u"Python Agent Test (agent_features)",
        u"entity.type": u"SERVICE",
        u"message": u"oops",
        u"log.level": u"ERROR",
        u"logger.name": u"test_logs_in_context",
        u"thread.name": u"MainThread",
        u"process.name": u"MainProcess",
        u"error.class": u"test_logs_in_context:ExceptionForTest",
        u"error.message": u"",
    }
    expected_extra_txn_keys = (
        "entity.guid",
        "hostname",
    )

    for k, v in expected.items():
        assert message.pop(k) == v

    assert set(message.keys()) == set(expected_extra_txn_keys)



EXPECTED_KEYS_TXN = (
    "trace.id",
    "span.id",
    "entity.name",
    "entity.type",
    "entity.guid",
        "hostname",
)

EXPECTED_KEYS_NO_TXN = EXPECTED_KEYS_TRACE_ENDED = (
    "entity.name",
    "entity.type",
    "entity.guid",
    "hostname",
)


def validate_metadata(metadata, expected):
    for key in expected:
        assert key in metadata, key
        assert metadata[key], metadata[key]
    for key in metadata.keys():
        assert key in expected, key


@background_task(name="test_get_linking_metadata_api")
def test_get_linking_metadata_api_inside_transaction():
    with FunctionTrace("test_linking_metadata") as trace:
        metadata = trace.get_linking_metadata()
        validate_metadata(metadata, EXPECTED_KEYS_TXN)
    metadata = trace.get_linking_metadata()
    validate_metadata(metadata, EXPECTED_KEYS_TRACE_ENDED)


def test_get_linking_metadata_api_outside_transaction():
    metadata = get_linking_metadata()
    validate_metadata(metadata, EXPECTED_KEYS_NO_TXN)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import psycopg2
import pytest

from testing_support.fixtures import override_application_settings
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.validators.validate_database_trace_inputs import validate_database_trace_inputs
from testing_support.util import instance_hostname
from utils import DB_MULTIPLE_SETTINGS
from testing_support.db_settings import postgresql_settings

from newrelic.api.background_task import background_task


# Settings

_enable_instance_settings = {
    'datastore_tracer.instance_reporting.enabled': True,
}
_disable_instance_settings = {
    'datastore_tracer.instance_reporting.enabled': False,
}


# Metrics

_base_scoped_metrics = [
        ('Function/psycopg2:connect', 2),
        ('Datastore/statement/Postgres/pg_settings/select', 1),
        ('Datastore/operation/Postgres/drop', 1),
        ('Datastore/operation/Postgres/create', 1),
        ('Datastore/operation/Postgres/commit', 2),
]

_base_rollup_metrics = [
        ('Datastore/all', 7),
        ('Datastore/allOther', 7),
        ('Datastore/Postgres/all', 7),
        ('Datastore/Postgres/allOther', 7),
        ('Datastore/statement/Postgres/pg_settings/select', 1),
        ('Datastore/operation/Postgres/drop', 1),
        ('Datastore/operation/Postgres/create', 1),
        ('Datastore/operation/Postgres/commit', 2),
]

_enable_scoped_metrics = list(_base_scoped_metrics)
_enable_rollup_metrics = list(_base_rollup_metrics)

_disable_scoped_metrics = list(_base_scoped_metrics)
_disable_rollup_metrics = list(_base_rollup_metrics)

if len(DB_MULTIPLE_SETTINGS) > 1:
    _postgresql_1 = DB_MULTIPLE_SETTINGS[0]
    _host_1 = instance_hostname(_postgresql_1['host'])
    _port_1 = _postgresql_1['port']

    _postgresql_2 = DB_MULTIPLE_SETTINGS[1]
    _host_2 = instance_hostname(_postgresql_2['host'])
    _port_2 = _postgresql_2['port']

    _instance_metric_name_1 = 'Datastore/instance/Postgres/%s/%s' % (
            _host_1, _port_1)
    _instance_metric_name_2 = 'Datastore/instance/Postgres/%s/%s' % (
            _host_2, _port_2)

    _enable_rollup_metrics.extend([
            (_instance_metric_name_1, 2),
            (_instance_metric_name_2, 3),
    ])
    _disable_rollup_metrics.extend([
            (_instance_metric_name_1, None),
            (_instance_metric_name_2, None),
    ])


# Query

def _exercise_db():

    postgresql1 = DB_MULTIPLE_SETTINGS[0]
    postgresql2 = DB_MULTIPLE_SETTINGS[1]

    connection = psycopg2.connect(
            database=postgresql1['name'], user=postgresql1['user'],
            password=postgresql1['password'], host=postgresql1['host'],
            port=postgresql1['port'])
    try:
        cursor = connection.cursor()
        cursor.execute("""SELECT setting from pg_settings where name=%s""",
                ('server_version',))
        connection.commit()
    finally:
        connection.close()

    connection = psycopg2.connect(
            database=postgresql2['name'], user=postgresql2['user'],
            password=postgresql2['password'], host=postgresql2['host'],
            port=postgresql2['port'])
    try:
        cursor = connection.cursor()
        cursor.execute("""drop table if exists %s""" % postgresql2["table_name"])
        cursor.execute("""create table %s """ % postgresql2["table_name"] +
                """(a integer, b real, c text)""")
        connection.commit()
    finally:
        connection.close()


# Tests

@pytest.mark.skipif(len(DB_MULTIPLE_SETTINGS) < 2,
        reason='Test environment not configured with multiple databases.')
@override_application_settings(_enable_instance_settings)
@validate_transaction_metrics(
        'test_multiple_dbs:test_multiple_databases_enable_instance',
        scoped_metrics=_enable_scoped_metrics,
        rollup_metrics=_enable_rollup_metrics,
        background_task=True)
@validate_database_trace_inputs(sql_parameters_type=tuple)
@background_task()
def test_multiple_databases_enable_instance():
    _exercise_db()

@pytest.mark.skipif(len(DB_MULTIPLE_SETTINGS) < 2,
        reason='Test environment not configured with multiple databases.')
@override_application_settings(_disable_instance_settings)
@validate_transaction_metrics(
        'test_multiple_dbs:test_multiple_databases_disable_instance',
        scoped_metrics=_disable_scoped_metrics,
        rollup_metrics=_disable_scoped_metrics,
        background_task=True)
@validate_database_trace_inputs(sql_parameters_type=tuple)
@background_task()
def test_multiple_databases_disable_instance():
    _exercise_db()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.common.object_wrapper import wrap_function_wrapper
from newrelic.api.transaction import current_transaction, record_log_event
from newrelic.core.config import global_settings
from newrelic.api.application import application_instance
from newrelic.hooks.logger_logging import add_nr_linking_metadata
from newrelic.common.signature import bind_args


def normalize_level_name(method_name):
    # Look up level number for method name, using result to look up level name for that level number.
    # Convert result to upper case, and default to UNKNOWN in case of errors or missing values.
    try:
        from structlog._log_levels import _LEVEL_TO_NAME, _NAME_TO_LEVEL
        return _LEVEL_TO_NAME[_NAME_TO_LEVEL[method_name]].upper()
    except Exception:
        return "UNKNOWN"


def bind_process_event(method_name, event, event_kw):
    return method_name, event, event_kw


def wrap__process_event(wrapped, instance, args, kwargs):
    try:
        method_name, event, event_kw = bind_process_event(*args, **kwargs)
    except TypeError:
        return wrapped(*args, **kwargs)

    original_message = event  # Save original undecorated message

    transaction = current_transaction()

    if transaction:
        settings = transaction.settings
    else:
        settings = global_settings()

    # Return early if application logging not enabled
    if settings and settings.application_logging and settings.application_logging.enabled:
        if settings.application_logging.local_decorating and settings.application_logging.local_decorating.enabled:
            event = add_nr_linking_metadata(event)

        # Send log to processors for filtering, allowing any DropEvent exceptions that occur to prevent instrumentation from recording the log event.
        result = wrapped(method_name, event, event_kw)
        
        level_name = normalize_level_name(method_name)

        if settings.application_logging.metrics and settings.application_logging.metrics.enabled:
            if transaction:
                transaction.record_custom_metric("Logging/lines", {"count": 1})
                transaction.record_custom_metric("Logging/lines/%s" % level_name, {"count": 1})
            else:
                application = application_instance(activate=False)
                if application and application.enabled:
                    application.record_custom_metric("Logging/lines", {"count": 1})
                    application.record_custom_metric("Logging/lines/%s" % level_name, {"count": 1})

        if settings.application_logging.forwarding and settings.application_logging.forwarding.enabled:
            try:
                record_log_event(original_message, level_name)

            except Exception:
                pass

        # Return the result from wrapped after we've recorded the resulting log event.
        return result

    return wrapped(*args, **kwargs)


def wrap__find_first_app_frame_and_name(wrapped, instance, args, kwargs):
    try:
        bound_args = bind_args(wrapped, args, kwargs)
        if bound_args["additional_ignores"]:
            bound_args["additional_ignores"] = list(bound_args["additional_ignores"])
            bound_args["additional_ignores"].append("newrelic")
        else:
            bound_args["additional_ignores"] = ["newrelic"]
    except Exception:
        return wrapped(*args, **kwargs)

    return wrapped(**bound_args)


def instrument_structlog__base(module):
    if hasattr(module, "BoundLoggerBase") and hasattr(module.BoundLoggerBase, "_process_event"):
        wrap_function_wrapper(module, "BoundLoggerBase._process_event", wrap__process_event)


def instrument_structlog__frames(module):
    if hasattr(module, "_find_first_app_frame_and_name"):
        wrap_function_wrapper(module, "_find_first_app_frame_and_name", wrap__find_first_app_frame_and_name)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import platform

from newrelic.api.application import application_settings
from newrelic.api.background_task import background_task
from testing_support.fixtures import reset_core_stats_engine
from testing_support.validators.validate_log_event_count import validate_log_event_count
from testing_support.fixtures import override_application_settings
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics

def get_metadata_string(log_message, is_txn):
    host = platform.uname().node
    assert host
    entity_guid = application_settings().entity_guid
    if is_txn:
        metadata_string = "".join(('NR-LINKING|', entity_guid, '|', host, '|abcdefgh12345678|abcdefgh|Python%20Agent%20Test%20%28internal_logging%29|'))
    else:
        metadata_string = "".join(('NR-LINKING|', entity_guid, '|', host, '|||Python%20Agent%20Test%20%28internal_logging%29|'))
    formatted_string = log_message + " " + metadata_string
    return formatted_string


def basic_logging(logger):
    logger.warning("C")


_settings_matrix = [
    (True, True, True),
    (True, False, False),
    (False, True, False),
    (False, False, False),
]


@pytest.mark.parametrize("feature_setting,subfeature_setting,expected", _settings_matrix)
@reset_core_stats_engine()
def test_log_forwarding_settings(logger, feature_setting, subfeature_setting, expected):
    @override_application_settings({
        "application_logging.enabled": feature_setting,
        "application_logging.forwarding.enabled": subfeature_setting,
    })
    @validate_log_event_count(1 if expected else 0)
    @background_task()
    def test():
        basic_logging(logger)
        assert len(logger.caplog.records) == 1

    test()


@pytest.mark.parametrize("feature_setting,subfeature_setting,expected", _settings_matrix)
@reset_core_stats_engine()
def test_local_decorating_settings(logger, feature_setting, subfeature_setting, expected):
    @override_application_settings({
        "application_logging.enabled": feature_setting,
        "application_logging.local_decorating.enabled": subfeature_setting,
    })
    @background_task()
    def test():
        basic_logging(logger)
        assert len(logger.caplog.records) == 1
        message = logger.caplog.records.pop()
        if expected:
            assert len(message) > 1
        else:
            assert len(message) == 1

    test()


@pytest.mark.parametrize("feature_setting,subfeature_setting,expected", _settings_matrix)
@reset_core_stats_engine()
def test_log_metrics_settings(logger, feature_setting, subfeature_setting, expected):
    metric_count = 1 if expected else None
    @override_application_settings({
        "application_logging.enabled": feature_setting,
        "application_logging.metrics.enabled": subfeature_setting,
    })
    @validate_transaction_metrics(
        "test_settings:test_log_metrics_settings.<locals>.test",
        custom_metrics=[
            ("Logging/lines", metric_count),
            ("Logging/lines/WARNING", metric_count),
        ],
        background_task=True,
    )
    @background_task()
    def test():
        basic_logging(logger)
        assert len(logger.caplog.records) == 1

    test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from testing_support.fixtures import collector_agent_registration_fixture, collector_available_fixture  # noqa: F401; pylint: disable=W0611

_default_settings = {
    'transaction_tracer.explain_threshold': 0.0,
    'transaction_tracer.transaction_threshold': 0.0,
    'transaction_tracer.stack_trace_threshold': 0.0,
    'debug.log_data_collector_payloads': True,
    'debug.record_transaction_failure': True,
}

collector_agent_registration = collector_agent_registration_fixture(
        app_name='Python Agent Test (framework_tornado)',
        default_settings=_default_settings)


@pytest.fixture(scope='module')
def app(request):
    import tornado
    from tornado.testing import AsyncHTTPTestCase
    from _target_application import make_app

    class App(AsyncHTTPTestCase):
        def get_app(self):
            custom = request.node.get_closest_marker("custom_app")
            return make_app(custom)

        def runTest(self, *args, **kwargs):
            pass

        @property
        def tornado_version(self):
            return '.'.join(map(str, tornado.version_info))

    case = App()
    case.setUp()
    yield case
    case.tearDown()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    from django.conf.urls.defaults import url
except ImportError:
    try:
        from django.conf.urls import url
    except ImportError:
        from django.urls import re_path as url

from rest_framework.decorators import api_view
from rest_framework.response import Response
from rest_framework.views import APIView
from rest_framework.settings import APISettings, api_settings
from views import index


class View(APIView):
    def get(self, request, format=None):
        return Response([{'message': 'restframework view response'}])


class Error(Exception):
    pass


class ViewError(APIView):
    def get(self, request, format=None):
        raise Error('xxx')


class ViewHandleError(APIView):
    settings = APISettings(api_settings.user_settings,
            api_settings.defaults, api_settings.import_strings)

    def get(self, request, status, global_exc):
        self.status = int(status)
        self.global_exc = global_exc == 'True'
        raise Error('omg cats')

    def _exception_handler(self, exc, context=None):
        if context:
            status = int(context['kwargs']['status'])
        else:
            status = self.status
        return Response([{'response': 'exception was handled global'}],
                status=status)

    def get_exception_handler(self):
        return self.settings.EXCEPTION_HANDLER

    def handle_exception(self, exc):
        self.settings.EXCEPTION_HANDLER = self._exception_handler
        if self.global_exc:
            return super(ViewHandleError, self).handle_exception(exc)
        else:
            return Response([{'response': 'exception was handled not global'}],
                    status=self.status)


@api_view(http_method_names=['GET'])
def wrapped_view(request):
    return Response({'message': 'wrapped_view response'})


urlpatterns = [
    url(r'^$', index, name='index'),
    url(r'^view/$', View.as_view()),
    url(r'^view_error/$', ViewError.as_view()),
    url(r'^view_handle_error/(?P<status>\d+)/(?P<global_exc>\w+)/$',
        ViewHandleError.as_view()),
    url(r'^api_view/$', wrapped_view),
]



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time

from newrelic.api.application import application_instance as application
from newrelic.api.background_task import background_task
from newrelic.api.transaction import record_custom_event
from newrelic.core.custom_event import process_event_type

from testing_support.fixtures import (reset_core_stats_engine,
        validate_custom_event_count,
        validate_custom_event_in_application_stats_engine,
        override_application_settings, function_not_called)

# Test process_event_type()

def test_process_event_type_name_is_string():
    name = 'string'
    assert process_event_type(name) == name

def test_process_event_type_name_is_not_string():
    name = 42
    assert process_event_type(name) is None

def test_process_event_type_name_ok_length():
    ok_name = 'CustomEventType'
    assert process_event_type(ok_name) == ok_name

def test_process_event_type_name_too_long():
    too_long = 'a' * 256
    assert process_event_type(too_long) is None

def test_process_event_type_name_valid_chars():
    valid_name = 'az09: '
    assert process_event_type(valid_name) == valid_name

def test_process_event_type_name_invalid_chars():
    invalid_name = '&'
    assert process_event_type(invalid_name) is None

_now = time.time()

_intrinsics = {
    'type': 'FooEvent',
    'timestamp': _now,
}
_user_params = {'foo': 'bar'}
_event = [_intrinsics, _user_params]

@reset_core_stats_engine()
@validate_custom_event_in_application_stats_engine(_event)
@background_task()
def test_add_custom_event_to_transaction_stats_engine():
    record_custom_event('FooEvent', _user_params)

@reset_core_stats_engine()
@validate_custom_event_in_application_stats_engine(_event)
def test_add_custom_event_to_application_stats_engine():
    app = application()
    record_custom_event('FooEvent', _user_params, application=app)

@reset_core_stats_engine()
@validate_custom_event_count(count=0)
@background_task()
def test_custom_event_inside_transaction_bad_event_type():
    record_custom_event('!@#$%^&*()', {'foo': 'bar'})

@reset_core_stats_engine()
@validate_custom_event_count(count=0)
@background_task()
def test_custom_event_outside_transaction_bad_event_type():
    app = application()
    record_custom_event('!@#$%^&*()', {'foo': 'bar'}, application=app)

_mixed_params = {'foo': 'bar', 123: 'bad key'}

@reset_core_stats_engine()
@validate_custom_event_in_application_stats_engine(_event)
@background_task()
def test_custom_event_inside_transaction_mixed_params():
    record_custom_event('FooEvent', _mixed_params)

@reset_core_stats_engine()
@validate_custom_event_in_application_stats_engine(_event)
@background_task()
def test_custom_event_outside_transaction_mixed_params():
    app = application()
    record_custom_event('FooEvent', _mixed_params, application=app)

_bad_params = {'*' * 256: 'too long', 123: 'bad key'}
_event_with_no_params = [{'type': 'FooEvent', 'timestamp': _now}, {}]

@reset_core_stats_engine()
@validate_custom_event_in_application_stats_engine(_event_with_no_params)
@background_task()
def test_custom_event_inside_transaction_bad_params():
    record_custom_event('FooEvent', _bad_params)

@reset_core_stats_engine()
@validate_custom_event_in_application_stats_engine(_event_with_no_params)
@background_task()
def test_custom_event_outside_transaction_bad_params():
    app = application()
    record_custom_event('FooEvent', _bad_params, application=app)

@reset_core_stats_engine()
@validate_custom_event_count(count=0)
@background_task()
def test_custom_event_params_not_a_dict():
    record_custom_event('ParamsListEvent', ['not', 'a', 'dict'])

# Tests for Custom Events configuration settings

@override_application_settings({'collect_custom_events': False})
@reset_core_stats_engine()
@validate_custom_event_count(count=0)
@background_task()
def test_custom_event_settings_check_collector_flag():
    record_custom_event('FooEvent', _user_params)

@override_application_settings({'custom_insights_events.enabled': False})
@reset_core_stats_engine()
@validate_custom_event_count(count=0)
@background_task()
def test_custom_event_settings_check_custom_insights_enabled():
    record_custom_event('FooEvent', _user_params)

# Test that record_custom_event() methods will short-circuit.
#
# If the custom_insights_events setting is False, verify that the
# `create_custom_event()` function is not called, in order to avoid the
# event_type and attribute processing.

@override_application_settings({'custom_insights_events.enabled': False})
@function_not_called('newrelic.api.transaction', 'create_custom_event')
@background_task()
def test_transaction_create_custom_event_not_called():
    record_custom_event('FooEvent', _user_params)

@override_application_settings({'custom_insights_events.enabled': False})
@function_not_called('newrelic.core.application', 'create_custom_event')
@background_task()
def test_application_create_custom_event_not_called():
    app = application()
    record_custom_event('FooEvent', _user_params, application=app)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from testing_support.fixtures import (
    override_application_settings,
    validate_tt_parenting)
from testing_support.validators.validate_code_level_metrics import validate_code_level_metrics
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.validators.validate_transaction_errors import validate_transaction_errors

from newrelic.packages import six

from conftest import async_handler_support, skip_if_not_async_handler_support

try:
    # The __version__ attribute was only added in 0.7.0.
    # Flask team does not use semantic versioning during development.
    from flask import __version__ as flask_version
    flask_version = tuple([int(v) for v in flask_version.split('.')])
    is_gt_flask060 = True
    is_dev_version = False
except ValueError:
    is_gt_flask060 = True
    is_dev_version = True
except ImportError:
    is_gt_flask060 = False
    is_dev_version = False

requires_endpoint_decorator = pytest.mark.skipif(not is_gt_flask060,
        reason="The endpoint decorator is not supported.")


def target_application():
    # We need to delay Flask application creation because of ordering
    # issues whereby the agent needs to be initialised before Flask is
    # imported and the routes configured. Normally pytest only runs the
    # global fixture which will initialise the agent after each test
    # file is imported, which is too late. We also can't do application
    # creation within a function as we will then get view handler
    # functions are different between Python 2 and 3, with the latter
    # showing <local> scope in path.

    if not async_handler_support:
        from _test_application import _test_application
    else:
        from _test_application_async import _test_application
    return _test_application


_test_application_index_scoped_metrics = [
        ('Function/flask.app:Flask.wsgi_app', 1),
        ('Python/WSGI/Application', 1),
        ('Python/WSGI/Response', 1),
        ('Python/WSGI/Finalize', 1),
        ('Function/_test_application:index_page', 1),
        ('Function/werkzeug.wsgi:ClosingIterator.close', 1)]

_test_application_index_tt_parenting = (
    'TransactionNode', [
        ('FunctionNode', [
            ('FunctionNode', [
                ('FunctionNode', []),
                ('FunctionNode', []),
                ('FunctionNode', []),
                # some flask versions have more FunctionNodes here, as appended
                # below
            ]),
        ]),
        ('FunctionNode', []),
        ('FunctionNode', [
            ('FunctionNode', []),
        ]),
    ]
)

if is_dev_version or (is_gt_flask060 and flask_version >= (0, 7)):
    _test_application_index_tt_parenting[1][0][1][0][1].append(
        ('FunctionNode', []),
    )
if is_dev_version or (is_gt_flask060 and flask_version >= (0, 9)):
    _test_application_index_tt_parenting[1][0][1][0][1].append(
        ('FunctionNode', []),
    )

@validate_transaction_errors(errors=[])
@validate_transaction_metrics('_test_application:index_page',
        scoped_metrics=_test_application_index_scoped_metrics)
@validate_tt_parenting(_test_application_index_tt_parenting)
@validate_code_level_metrics("_test_application", "index_page")
def test_application_index():
    application = target_application()
    response = application.get('/index')
    response.mustcontain('INDEX RESPONSE')

_test_application_async_scoped_metrics = [
        ('Function/flask.app:Flask.wsgi_app', 1),
        ('Python/WSGI/Application', 1),
        ('Python/WSGI/Response', 1),
        ('Python/WSGI/Finalize', 1),
        ('Function/_test_application_async:async_page', 1),
        ('Function/werkzeug.wsgi:ClosingIterator.close', 1)]

@skip_if_not_async_handler_support
@validate_transaction_errors(errors=[])
@validate_transaction_metrics('_test_application_async:async_page',
        scoped_metrics=_test_application_async_scoped_metrics)
@validate_tt_parenting(_test_application_index_tt_parenting)
@validate_code_level_metrics("_test_application_async", "async_page")
def test_application_async():
    application = target_application()
    response = application.get('/async')
    response.mustcontain('ASYNC RESPONSE')

_test_application_endpoint_scoped_metrics = [
        ('Function/flask.app:Flask.wsgi_app', 1),
        ('Python/WSGI/Application', 1),
        ('Python/WSGI/Response', 1),
        ('Python/WSGI/Finalize', 1),
        ('Function/_test_application:endpoint_page', 1),
        ('Function/werkzeug.wsgi:ClosingIterator.close', 1)]


@validate_transaction_errors(errors=[])
@validate_transaction_metrics('_test_application:endpoint_page',
        scoped_metrics=_test_application_endpoint_scoped_metrics)
@validate_code_level_metrics("_test_application", "endpoint_page")
def test_application_endpoint():
    application = target_application()
    response = application.get('/endpoint')
    response.mustcontain('ENDPOINT RESPONSE')


_test_application_error_scoped_metrics = [
        ('Function/flask.app:Flask.wsgi_app', 1),
        ('Python/WSGI/Application', 1),
        ('Python/WSGI/Response', 1),
        ('Python/WSGI/Finalize', 1),
        ('Function/_test_application:error_page', 1),
        ('Function/flask.app:Flask.handle_exception', 1),
        ('Function/werkzeug.wsgi:ClosingIterator.close', 1),
        ('Function/flask.app:Flask.handle_user_exception', 1),
        ('Function/flask.app:Flask.handle_user_exception', 1)]


if six.PY3:
    _test_application_error_errors = ['builtins:RuntimeError']
else:
    _test_application_error_errors = ['exceptions:RuntimeError']


@validate_transaction_errors(errors=_test_application_error_errors)
@validate_transaction_metrics('_test_application:error_page',
        scoped_metrics=_test_application_error_scoped_metrics)
@validate_code_level_metrics("_test_application", "error_page")
def test_application_error():
    application = target_application()
    application.get('/error', status=500, expect_errors=True)


_test_application_abort_404_scoped_metrics = [
        ('Function/flask.app:Flask.wsgi_app', 1),
        ('Python/WSGI/Application', 1),
        ('Python/WSGI/Response', 1),
        ('Python/WSGI/Finalize', 1),
        ('Function/_test_application:abort_404_page', 1),
        ('Function/flask.app:Flask.handle_http_exception', 1),
        ('Function/werkzeug.wsgi:ClosingIterator.close', 1),
        ('Function/flask.app:Flask.handle_user_exception', 1)]


@validate_transaction_errors(errors=[])
@validate_transaction_metrics('_test_application:abort_404_page',
        scoped_metrics=_test_application_abort_404_scoped_metrics)
@validate_code_level_metrics("_test_application", "abort_404_page")
def test_application_abort_404():
    application = target_application()
    application.get('/abort_404', status=404)


_test_application_exception_404_scoped_metrics = [
        ('Function/flask.app:Flask.wsgi_app', 1),
        ('Python/WSGI/Application', 1),
        ('Python/WSGI/Response', 1),
        ('Python/WSGI/Finalize', 1),
        ('Function/_test_application:exception_404_page', 1),
        ('Function/flask.app:Flask.handle_http_exception', 1),
        ('Function/werkzeug.wsgi:ClosingIterator.close', 1),
        ('Function/flask.app:Flask.handle_user_exception', 1)]


@validate_transaction_errors(errors=[])
@validate_transaction_metrics('_test_application:exception_404_page',
        scoped_metrics=_test_application_exception_404_scoped_metrics)
@validate_code_level_metrics("_test_application", "exception_404_page")
def test_application_exception_404():
    application = target_application()
    application.get('/exception_404', status=404)


_test_application_not_found_scoped_metrics = [
        ('Function/flask.app:Flask.wsgi_app', 1),
        ('Python/WSGI/Application', 1),
        ('Python/WSGI/Response', 1),
        ('Python/WSGI/Finalize', 1),
        ('Function/flask.app:Flask.handle_http_exception', 1),
        ('Function/werkzeug.wsgi:ClosingIterator.close', 1),
        ('Function/flask.app:Flask.handle_user_exception', 1)]


@validate_transaction_errors(errors=[])
@validate_transaction_metrics('flask.app:Flask.handle_http_exception',
        scoped_metrics=_test_application_not_found_scoped_metrics)
def test_application_not_found():
    application = target_application()
    application.get('/missing', status=404)


_test_application_render_template_string_scoped_metrics = [
        ('Function/flask.app:Flask.wsgi_app', 1),
        ('Python/WSGI/Application', 1),
        ('Python/WSGI/Response', 1),
        ('Python/WSGI/Finalize', 1),
        ('Function/_test_application:template_string', 1),
        ('Function/werkzeug.wsgi:ClosingIterator.close', 1),
        ('Template/Compile/<template>', 1),
        ('Template/Render/<template>', 1)]


@validate_transaction_errors(errors=[])
@validate_transaction_metrics('_test_application:template_string',
        scoped_metrics=_test_application_render_template_string_scoped_metrics)
@validate_code_level_metrics("_test_application", "template_string")
def test_application_render_template_string():
    application = target_application()
    application.get('/template_string')


_test_application_render_template_not_found_scoped_metrics = [
        ('Function/flask.app:Flask.wsgi_app', 1),
        ('Python/WSGI/Application', 1),
        ('Python/WSGI/Response', 1),
        ('Python/WSGI/Finalize', 1),
        ('Function/_test_application:template_not_found', 1),
        ('Function/flask.app:Flask.handle_exception', 1),
        ('Function/werkzeug.wsgi:ClosingIterator.close', 1),
        ('Function/flask.app:Flask.handle_user_exception', 1)]


@validate_transaction_errors(errors=['jinja2.exceptions:TemplateNotFound'])
@validate_transaction_metrics('_test_application:template_not_found',
    scoped_metrics=_test_application_render_template_not_found_scoped_metrics)
def test_application_render_template_not_found():
    application = target_application()
    application.get('/template_not_found', status=500, expect_errors=True)


_test_html_insertion_settings = {
    'browser_monitoring.enabled': True,
    'browser_monitoring.auto_instrument': True,
    'js_agent_loader': u'<!-- NREUM HEADER -->',
}


@override_application_settings(_test_html_insertion_settings)
def test_html_insertion():
    application = target_application()
    response = application.get('/html_insertion', status=200)

    # The 'NREUM HEADER' value comes from our override for the header.
    # The 'NREUM.info' value comes from the programmatically generated
    # footer added by the agent.

    response.mustcontain('NREUM HEADER', 'NREUM.info')



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from testing_support.validators.validate_database_duration import (
    validate_database_duration,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)
from testing_support.validators.validate_tt_collector_json import (
    validate_tt_collector_json,
)

from newrelic.api.background_task import background_task


@pytest.fixture()
def exercise_async_collections(async_collection):
    async def _exercise_async_collections():
        async_collection.document("DoesNotExist")
        await async_collection.add({"capital": "Rome", "currency": "Euro", "language": "Italian"}, "Italy")
        await async_collection.add({"capital": "Mexico City", "currency": "Peso", "language": "Spanish"}, "Mexico")

        documents_get = await async_collection.get()
        assert len(documents_get) == 2
        documents_stream = [_ async for _ in async_collection.stream()]
        assert len(documents_stream) == 2
        documents_list = [_ async for _ in async_collection.list_documents()]
        assert len(documents_list) == 2

    return _exercise_async_collections


def test_firestore_async_collections(loop, exercise_async_collections, async_collection, instance_info):
    _test_scoped_metrics = [
        ("Datastore/statement/Firestore/%s/stream" % async_collection.id, 1),
        ("Datastore/statement/Firestore/%s/get" % async_collection.id, 1),
        ("Datastore/statement/Firestore/%s/list_documents" % async_collection.id, 1),
        ("Datastore/statement/Firestore/%s/add" % async_collection.id, 2),
    ]

    _test_rollup_metrics = [
        ("Datastore/operation/Firestore/add", 2),
        ("Datastore/operation/Firestore/get", 1),
        ("Datastore/operation/Firestore/stream", 1),
        ("Datastore/operation/Firestore/list_documents", 1),
        ("Datastore/all", 5),
        ("Datastore/allOther", 5),
        ("Datastore/instance/Firestore/%s/%s" % (instance_info["host"], instance_info["port_path_or_id"]), 5),
    ]

    @validate_database_duration()
    @validate_transaction_metrics(
        "test_firestore_async_collections",
        scoped_metrics=_test_scoped_metrics,
        rollup_metrics=_test_rollup_metrics,
        background_task=True,
    )
    @background_task(name="test_firestore_async_collections")
    def _test():
        loop.run_until_complete(exercise_async_collections())

    _test()


@background_task()
def test_firestore_async_collections_generators(collection, async_collection, assert_trace_for_async_generator):
    collection.add({})
    collection.add({})
    assert len([_ for _ in collection.list_documents()]) == 2

    assert_trace_for_async_generator(async_collection.stream)
    assert_trace_for_async_generator(async_collection.list_documents)


def test_firestore_async_collections_trace_node_datastore_params(loop, exercise_async_collections, instance_info):
    @validate_tt_collector_json(datastore_params=instance_info)
    @background_task()
    def _test():
        loop.run_until_complete(exercise_async_collections())

    _test()


# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sqlite3

from testing_support.validators.validate_database_duration import (
    validate_database_duration,
)

from newrelic.api.background_task import background_task

from conftest import ES_VERSION




def _exercise_es_v7(es):
    es.index(index="contacts", doc_type="person", body={"name": "Joe Tester", "age": 25, "title": "QA Engineer"}, id=1)
    es.index(
        index="contacts", doc_type="person", body={"name": "Jessica Coder", "age": 32, "title": "Programmer"}, id=2
    )
    es.index(index="contacts", doc_type="person", body={"name": "Freddy Tester", "age": 29, "title": "Assistant"}, id=3)
    es.indices.refresh("contacts")


def _exercise_es_v8(es):
    es.index(index="contacts", body={"name": "Joe Tester", "age": 25, "title": "QA Engineer"}, id=1)
    es.index(index="contacts", body={"name": "Jessica Coder", "age": 32, "title": "Programmer"}, id=2)
    es.index(index="contacts", body={"name": "Freddy Tester", "age": 29, "title": "Assistant"}, id=3)
    es.indices.refresh(index="contacts")


_exercise_es = _exercise_es_v7 if ES_VERSION < (8, 0, 0) else _exercise_es_v8


@validate_database_duration()
@background_task()
def test_elasticsearch_database_duration(client):
    _exercise_es(client)


@validate_database_duration()
@background_task()
def test_elasticsearch_and_sqlite_database_duration(client):

    # Make Elasticsearch queries

    _exercise_es(client)

    # Make sqlite queries

    conn = sqlite3.connect(":memory:")
    cur = conn.cursor()

    cur.execute("CREATE TABLE people (name text, age int)")
    cur.execute("INSERT INTO people VALUES ('Bob', 22)")

    conn.commit()
    conn.close()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time

import kafka
from testing_support.validators.validate_custom_metrics_outside_transaction import (
    validate_custom_metrics_outside_transaction,
)


@validate_custom_metrics_outside_transaction(
    [
        ("MessageBroker/Kafka/Heartbeat/Poll", "present"),
        ("MessageBroker/Kafka/Heartbeat/Sent", "present"),
        ("MessageBroker/Kafka/Heartbeat/Receive", "present"),
        ("MessageBroker/Kafka/Heartbeat/Fail", None),
        ("MessageBroker/Kafka/Heartbeat/SessionTimeout", None),
        ("MessageBroker/Kafka/Heartbeat/PollTimeout", None),
    ]
)
def test_successful_heartbeat_metrics_recorded(topic, get_consumer_record):
    get_consumer_record()
    time.sleep(1.5)


@validate_custom_metrics_outside_transaction(
    [
        ("MessageBroker/Kafka/Heartbeat/Poll", "present"),
        ("MessageBroker/Kafka/Heartbeat/Sent", "present"),
        ("MessageBroker/Kafka/Heartbeat/Fail", "present"),
        ("MessageBroker/Kafka/Heartbeat/Receive", "present"),
        ("MessageBroker/Kafka/Heartbeat/SessionTimeout", "present"),
        ("MessageBroker/Kafka/Heartbeat/PollTimeout", "present"),
    ]
)
def test_fail_timeout_heartbeat_metrics_recorded():
    heartbeat = kafka.coordinator.heartbeat.Heartbeat(session_timeout_ms=0, max_poll_interval_ms=0)

    heartbeat.poll()
    heartbeat.sent_heartbeat()
    heartbeat.received_heartbeat()
    heartbeat.fail_heartbeat()

    assert heartbeat.session_timeout_expired(), "Failed to force heartbeat to timeout."
    assert heartbeat.poll_timeout_expired(), "Failed to force heartbeat to timeout."



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import webtest

from flask import Flask, render_template_string, render_template, abort
from werkzeug.exceptions import NotFound
from werkzeug.routing import Rule

application = Flask(__name__)

@application.route('/index')
def index_page():
    return 'INDEX RESPONSE'

application.url_map.add(Rule('/endpoint', endpoint='endpoint'))

@application.endpoint('endpoint')
def endpoint_page():
    return 'ENDPOINT RESPONSE'

@application.route('/error')
def error_page():
    raise RuntimeError('RUNTIME ERROR')

@application.route('/abort_404')
def abort_404_page():
    abort(404)

@application.route('/exception_404')
def exception_404_page():
    raise NotFound()

@application.route('/template_string')
def template_string():
    return render_template_string('<body><p>INDEX RESPONSE</p></body>')

@application.route('/template_not_found')
def template_not_found():
    return render_template('not_found')

@application.route('/html_insertion')
def html_insertion():
    return ('<!DOCTYPE html><html><head>Some header</head>'
            '<body><h1>My First Heading</h1><p>My first paragraph.</p>'
            '</body></html>')

_test_application = webtest.TestApp(application)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tempfile
import os
import shutil
import sys
import pytest
import newrelic.core.agent as agent


@pytest.mark.parametrize('content', [{}, {'opt': [1, 2, 3]}])
def test_check_environment_failing(content):
    temp_dir = tempfile.mkdtemp()

    try:
        uwsgi_dir = os.path.join(temp_dir, 'uwsgi')
        init_file = os.path.join(uwsgi_dir, '__init__.py')
        os.makedirs(uwsgi_dir)
        with open(init_file, 'w') as f:
            for key, value in content.items():
                f.write("%s = %s" % (key, value))

        sys.path.insert(0, temp_dir)
        import uwsgi
        for key, value in content.items():
            assert getattr(uwsgi, key) == value

        agent.check_environment()
    finally:
        shutil.rmtree(temp_dir)
        sys.path.remove(temp_dir)
        del sys.modules['uwsgi']



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from newrelic.common.signature import bind_args


@pytest.mark.parametrize(
    "func,args,kwargs,expected",
    [
        (lambda x, y: None, (1,), {"y": 2}, {"x": 1, "y": 2}),
        (lambda x=1, y=2: None, (1,), {"y": 2}, {"x": 1, "y": 2}),
        (lambda x=1: None, (), {}, {"x": 1}),
    ],
    ids=("posargs", "kwargs", "defaults"),
)
def test_signature_binding(func, args, kwargs, expected):
    bound_args = bind_args(func, args, kwargs)
    assert bound_args == expected



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import webtest
from testing_support.sample_applications import simple_app
from testing_support.validators.validate_apdex_metrics import validate_apdex_metrics

from newrelic.api.transaction import current_transaction, suppress_apdex_metric
from newrelic.api.wsgi_application import wsgi_application

normal_application = webtest.TestApp(simple_app)

# NOTE: This test validates that the server-side apdex_t is set to 0.5
# If the server-side configuration changes, this test will start to fail.


@validate_apdex_metrics(
    name="",
    group="Uri",
    apdex_t_min=0.5,
    apdex_t_max=0.5,
)
def test_apdex():
    normal_application.get("/")


# This has to be a Web Transaction.
# The apdex measurement only applies to Web Transactions
def test_apdex_suppression():
    @wsgi_application()
    def simple_apdex_supression_app(environ, start_response):
        suppress_apdex_metric()

        start_response(status="200 OK", response_headers=[])
        transaction = current_transaction()

        assert transaction.suppress_apdex
        assert transaction.apdex == 0
        return []

    apdex_suppression_app = webtest.TestApp(simple_apdex_supression_app)
    apdex_suppression_app.get("/")



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio

import pytest
from testing_support.db_settings import redis_settings
from testing_support.fixture.event_loop import event_loop as loop  # noqa: F401
from testing_support.util import instance_hostname
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.background_task import background_task
from newrelic.common.package_version_utils import get_package_version_tuple

# Settings

DB_SETTINGS = redis_settings()[0]
REDIS_PY_VERSION = get_package_version_tuple("redis")

# Metrics for publish test

datastore_all_metric_count = 5 if REDIS_PY_VERSION >= (5, 0) else 3

_base_scoped_metrics = [("Datastore/operation/Redis/publish", 3)]

if REDIS_PY_VERSION >= (5, 0):
    _base_scoped_metrics.append(
        ("Datastore/operation/Redis/client_setinfo", 2),
    )

_base_rollup_metrics = [
    ("Datastore/all", datastore_all_metric_count),
    ("Datastore/allOther", datastore_all_metric_count),
    ("Datastore/Redis/all", datastore_all_metric_count),
    ("Datastore/Redis/allOther", datastore_all_metric_count),
    ("Datastore/operation/Redis/publish", 3),
    (
        "Datastore/instance/Redis/%s/%s" % (instance_hostname(DB_SETTINGS["host"]), DB_SETTINGS["port"]),
        datastore_all_metric_count,
    ),
]
if REDIS_PY_VERSION >= (5, 0):
    _base_rollup_metrics.append(
        ("Datastore/operation/Redis/client_setinfo", 2),
    )


# Metrics for connection pool test

_base_pool_scoped_metrics = [
    ("Datastore/operation/Redis/get", 1),
    ("Datastore/operation/Redis/set", 1),
    ("Datastore/operation/Redis/client_list", 1),
]

_base_pool_rollup_metrics = [
    ("Datastore/all", 3),
    ("Datastore/allOther", 3),
    ("Datastore/Redis/all", 3),
    ("Datastore/Redis/allOther", 3),
    ("Datastore/operation/Redis/get", 1),
    ("Datastore/operation/Redis/set", 1),
    ("Datastore/operation/Redis/client_list", 1),
    ("Datastore/instance/Redis/%s/%s" % (instance_hostname(DB_SETTINGS["host"]), DB_SETTINGS["port"]), 3),
]


# Tests


@pytest.fixture()
def client(loop):  # noqa
    import redis.asyncio

    return loop.run_until_complete(redis.asyncio.Redis(host=DB_SETTINGS["host"], port=DB_SETTINGS["port"], db=0))


@pytest.fixture()
def client_pool(loop):  # noqa
    import redis.asyncio

    connection_pool = redis.asyncio.ConnectionPool(host=DB_SETTINGS["host"], port=DB_SETTINGS["port"], db=0)
    return loop.run_until_complete(redis.asyncio.Redis(connection_pool=connection_pool))


@pytest.mark.skipif(REDIS_PY_VERSION < (4, 2), reason="This functionality exists in Redis 4.2+")
@validate_transaction_metrics(
    "test_asyncio:test_async_connection_pool",
    scoped_metrics=_base_pool_scoped_metrics,
    rollup_metrics=_base_pool_rollup_metrics,
    background_task=True,
)
@background_task()
def test_async_connection_pool(client_pool, loop):  # noqa
    async def _test_async_pool(client_pool):
        await client_pool.set("key1", "value1")
        await client_pool.get("key1")
        await client_pool.execute_command("CLIENT", "LIST")

    loop.run_until_complete(_test_async_pool(client_pool))


@pytest.mark.skipif(REDIS_PY_VERSION < (4, 2), reason="This functionality exists in Redis 4.2+")
@validate_transaction_metrics("test_asyncio:test_async_pipeline", background_task=True)
@background_task()
def test_async_pipeline(client, loop):  # noqa
    async def _test_pipeline(client):
        async with client.pipeline(transaction=True) as pipe:
            await pipe.set("key1", "value1")
            await pipe.execute()

    loop.run_until_complete(_test_pipeline(client))


@pytest.mark.skipif(REDIS_PY_VERSION < (4, 2), reason="This functionality exists in Redis 4.2+")
@validate_transaction_metrics(
    "test_asyncio:test_async_pubsub",
    scoped_metrics=_base_scoped_metrics,
    rollup_metrics=_base_rollup_metrics,
    background_task=True,
)
@background_task()
def test_async_pubsub(client, loop):  # noqa
    messages_received = []

    async def reader(pubsub):
        while True:
            message = await pubsub.get_message(ignore_subscribe_messages=True)
            if message:
                messages_received.append(message["data"].decode())
                if message["data"].decode() == "NOPE":
                    break

    async def _test_pubsub():
        async with client.pubsub() as pubsub:
            await pubsub.psubscribe("channel:*")

            future = asyncio.create_task(reader(pubsub))

            await client.publish("channel:1", "Hello")
            await client.publish("channel:2", "World")
            await client.publish("channel:1", "NOPE")

            await future

    loop.run_until_complete(_test_pubsub())
    assert messages_received == ["Hello", "World", "NOPE"]



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from concurrent import futures
import grpc
import socket


# This defines an external grpc server test apps can use for testing.
#
# Example:
#
#   with MockExternalgRPCServer(port=PORT) as server:
#       add_SampleApplicationServicer_to_server(SampleApplicationServicer(),
#               server)
#       ... test stuff ...


class MockExternalgRPCServer(object):
    def __init__(self, port=None, *args, **kwargs):
        self.server = grpc.server(futures.ThreadPoolExecutor(max_workers=4))
        if port:
            self.server.port = self.server.add_insecure_port('127.0.0.1:%s' % port)
            self.port = port
        else:
            # If port not set, try to bind to a port until successful
            retries = 5  # Set retry limit to prevent infinite loops
            self.port = None  # Initialize empty
            while not self.port and retries > 0:
                retries -= 1
                try:
                    # Obtain random open port
                    port = self.get_open_port()
                    # Attempt to bind to port
                    self.server.port = self.server.add_insecure_port('127.0.0.1:%s' % port)
                    self.port = port
                except OSError as exc:
                    # Reraise errors other than port already in use
                    if "Address already in use" not in exc:
                        raise

    @staticmethod
    def get_open_port():
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.bind(("", 0))
        port = s.getsockname()[1]
        s.close()
        return port

    def __enter__(self):
        self.server.start()
        return self.server

    def __exit__(self, type, value, tb):
        # Set grace period to None so that the server shuts down immediately
        # when the context manager exits. This will hopefully prevent tests
        # from hanging while waiting for the server to shut down.
        # https://github.com/grpc/grpc/blob/662ec97674dd0918f4db4c21f5f47038c535a9ba/src/python/grpcio/grpc/_server.py#L729
        self.server.stop(grace=None)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.api.datastore_trace import DatastoreTrace
from newrelic.api.transaction import current_transaction
from newrelic.common.object_wrapper import function_wrapper, wrap_function_wrapper
from newrelic.common.package_version_utils import get_package_version_tuple
from newrelic.packages import six

# An index name can be a string, None or a sequence. In the case of None
# an empty string or '*', it is the same as using '_all'. When a string
# it can also be a comma separated list of index names. A sequence
# obviously can also be more than one index name. Where we are certain
# there is only a single index name we use it, otherwise we use 'other'.

ES_VERSION = get_package_version_tuple("elasticsearch")


def _index_name(index):
    if not index or index == "*":
        return "_all"
    if not isinstance(index, six.string_types) or "," in index:
        return "other"
    return index


def _extract_args_index(index=None, *args, **kwargs):
    return _index_name(index)


def _extract_args_allocation_explain_index(
    current_node=None,
    error_trace=None,
    filter_path=None,
    human=None,
    include_disk_info=None,
    include_yes_decisions=None,
    index=None,
    *args,
    **kwargs
):
    return _index_name(index)


def _extract_args_name_index(name=None, index=None, *args, **kwargs):
    return _index_name(index)


def _extract_args_body_index(body=None, index=None, *args, **kwargs):
    return _index_name(index)


def _extract_args_requests_index(requests=None, index=None, *args, **kwargs):
    return _index_name(index)


def _extract_args_searches_index(searches=None, index=None, *args, **kwargs):
    return _index_name(index)


def _extract_args_search_templates_index(search_templates=None, index=None, *args, **kwargs):
    return _index_name(index)


def _extract_args_operations_index(operations=None, index=None, *args, **kwargs):
    return _index_name(index)


def _extract_args_doctype_body_index(doc_type=None, body=None, index=None, *args, **kwargs):
    return _index_name(index)


def _extract_args_field_index(field=None, index=None, *args, **kwargs):
    return _index_name(index)


def _extract_args_fields_index(fields=None, index=None, *args, **kwargs):
    return _index_name(index)


def _extract_args_name_body_index(name=None, body=None, index=None, *args, **kwargs):
    return _index_name(index)


def _extract_args_metric_index(metric=None, index=None, *args, **kwargs):
    return _index_name(index)


def _extract_args_settings_index(settings=None, index=None, *args, **kwargs):
    return _index_name(index)


def instrument_es_methods(module, _class, client_methods, prefix=None):
    for method_name, arg_extractor in client_methods:
        if hasattr(getattr(module, _class), method_name):
            wrap_elasticsearch_client_method(module, _class, method_name, arg_extractor, prefix)


def wrap_elasticsearch_client_method(module, class_name, method_name, arg_extractor, prefix=None):
    def _nr_wrapper_Elasticsearch_method_(wrapped, instance, args, kwargs):
        transaction = current_transaction()

        if transaction is None:
            return wrapped(*args, **kwargs)
        # When index is None, it means there is no target field
        # associated with this method. Hence this method will only
        # create an operation metric and no statement metric. This is
        # handled by setting the target to None when calling the
        # DatastoreTraceWrapper.
        if arg_extractor is None:
            index = None
        else:
            index = arg_extractor(*args, **kwargs)

        if prefix:
            operation = "%s.%s" % (prefix, method_name)
        else:
            operation = method_name

        transaction._nr_datastore_instance_info = (None, None, None)

        dt = DatastoreTrace(product="Elasticsearch", target=index, operation=operation, source=wrapped)

        with dt:
            result = wrapped(*args, **kwargs)

            instance_info = transaction._nr_datastore_instance_info
            host, port_path_or_id, _ = instance_info

            dt.host = host
            dt.port_path_or_id = port_path_or_id

            return result

    wrap_function_wrapper(module, "%s.%s" % (class_name, method_name), _nr_wrapper_Elasticsearch_method_)


_elasticsearch_client_methods_below_v8 = (
    ("abort_benchmark", None),
    ("benchmark", _extract_args_index),
    ("bulk", None),
    ("clear_scroll", None),
    ("count", _extract_args_index),
    ("count_percolate", _extract_args_index),
    ("create", _extract_args_index),
    ("delete", _extract_args_index),
    ("delete_by_query", _extract_args_index),
    ("delete_script", None),
    ("delete_template", None),
    ("exists", _extract_args_index),
    ("explain", _extract_args_index),
    ("get", _extract_args_index),
    ("get_script", None),
    ("get_source", _extract_args_index),
    ("get_template", None),
    ("index", _extract_args_index),
    ("info", None),
    ("list_benchmarks", _extract_args_index),
    ("mget", None),
    ("mlt", _extract_args_index),
    ("mpercolate", _extract_args_body_index),
    ("msearch", None),
    ("mtermvectors", None),
    ("percolate", _extract_args_index),
    ("ping", None),
    ("put_script", None),
    ("put_template", None),
    ("scroll", None),
    ("search", _extract_args_index),
    ("search_exists", _extract_args_index),
    ("search_shards", _extract_args_index),
    ("search_template", _extract_args_index),
    ("suggest", _extract_args_body_index),
    ("termvector", _extract_args_index),
    ("termvectors", None),
    ("update", _extract_args_index),
)


_elasticsearch_client_methods_v8 = (
    ("bulk", _extract_args_operations_index),
    ("clear_scroll", None),
    ("close", None),
    ("close_point_in_time", None),
    ("count", _extract_args_index),
    ("create", _extract_args_index),
    ("delete", _extract_args_index),
    ("delete_by_query", _extract_args_index),
    ("delete_by_query_rethrottle", None),
    ("delete_script", None),
    ("exists", _extract_args_index),
    ("exists_source", _extract_args_index),
    ("explain", _extract_args_index),
    ("field_caps", _extract_args_index),
    ("get", _extract_args_index),
    ("get_script", None),
    ("get_script_context", None),
    ("get_script_languages", None),
    ("get_source", _extract_args_index),
    ("index", _extract_args_index),
    ("info", None),
    ("knn_search", _extract_args_index),
    ("mget", _extract_args_index),
    ("msearch", _extract_args_searches_index),
    ("msearch_template", _extract_args_search_templates_index),
    ("mtermvectors", _extract_args_index),
    ("open_point_in_time", _extract_args_index),
    ("options", None),
    ("ping", None),
    ("put_script", None),
    ("rank_eval", _extract_args_requests_index),
    ("reindex", None),
    ("reindex_rethrottle", None),
    ("render_search_template", None),
    ("scripts_painless_execute", None),
    ("scroll", None),
    ("search", _extract_args_index),
    ("search_mvt", _extract_args_index),
    ("search_shards", _extract_args_index),
    ("terms_enum", _extract_args_index),
    ("termvector", _extract_args_index),
    ("termvectors", _extract_args_index),
    ("update", _extract_args_index),
    ("update_by_query", _extract_args_index),
    ("update_by_query_rethrottle", None),
)


def instrument_elasticsearch_client(module):
    # The module path was remapped in v8 to match previous versions.
    # In order to avoid double wrapping we check the version before
    # wrapping.
    if ES_VERSION < (8,):
        instrument_es_methods(module, "Elasticsearch", _elasticsearch_client_methods_below_v8)


def instrument_elasticsearch_client_v8(module):
    instrument_es_methods(module, "Elasticsearch", _elasticsearch_client_methods_v8)


_elasticsearch_client_indices_methods_below_v8 = (
    ("analyze", _extract_args_index),
    ("clear_cache", _extract_args_index),
    ("close", _extract_args_index),
    ("create", _extract_args_index),
    ("delete", _extract_args_index),
    ("delete_alias", _extract_args_index),
    ("delete_mapping", _extract_args_index),
    ("delete_template", None),
    ("delete_warmer", _extract_args_index),
    ("exists", _extract_args_index),
    ("exists_alias", _extract_args_name_index),
    ("exists_template", None),
    ("exists_type", _extract_args_index),
    ("flush", _extract_args_index),
    ("get", _extract_args_index),
    ("get_alias", _extract_args_index),
    ("get_aliases", _extract_args_index),
    ("get_mapping", _extract_args_index),
    ("get_field_mapping", _extract_args_field_index),
    ("get_settings", _extract_args_index),
    ("get_template", None),
    ("get_upgrade", _extract_args_index),
    ("get_warmer", _extract_args_index),
    ("open", _extract_args_index),
    ("optimize", _extract_args_index),
    ("put_alias", _extract_args_name_index),
    ("put_mapping", _extract_args_doctype_body_index),
    ("put_settings", _extract_args_body_index),
    ("put_template", None),
    ("put_warmer", _extract_args_name_body_index),
    ("recovery", _extract_args_index),
    ("refresh", _extract_args_index),
    ("segments", _extract_args_index),
    ("snapshot_index", _extract_args_index),
    ("stats", _extract_args_index),
    ("status", _extract_args_index),
    ("update_aliases", None),
    ("upgrade", _extract_args_index),
    ("validate_query", _extract_args_index),
)


_elasticsearch_client_indices_methods_v8 = (
    ("add_block", _extract_args_index),
    ("analyze", _extract_args_index),
    ("clear_cache", _extract_args_index),
    ("clone", _extract_args_index),
    ("close", _extract_args_index),
    ("create", _extract_args_index),
    ("create_data_stream", None),
    ("data_streams_stats", None),
    ("delete", _extract_args_index),
    ("delete_alias", _extract_args_index),
    ("delete_data_stream", None),
    ("delete_index_template", None),
    ("delete_template", None),
    ("disk_usage", _extract_args_index),
    ("downsample", _extract_args_index),
    ("exists", _extract_args_index),
    ("exists_alias", _extract_args_name_index),
    ("exists_index_template", None),
    ("exists_template", None),
    ("field_usage_stats", _extract_args_index),
    ("flush", _extract_args_index),
    ("forcemerge", _extract_args_index),
    ("get", _extract_args_index),
    ("get_alias", _extract_args_index),
    ("get_data_stream", None),
    ("get_field_mapping", _extract_args_fields_index),
    ("get_index_template", None),
    ("get_mapping", _extract_args_index),
    ("get_settings", _extract_args_index),
    ("get_template", None),
    ("migrate_to_data_stream", None),
    ("modify_data_stream", None),
    ("open", _extract_args_index),
    ("promote_data_stream", None),
    ("put_alias", _extract_args_index),
    ("put_index_template", None),
    ("put_mapping", _extract_args_index),
    ("put_settings", _extract_args_settings_index),
    ("put_template", None),
    ("recovery", _extract_args_index),
    ("refresh", _extract_args_index),
    ("reload_search_analyzers", _extract_args_index),
    ("resolve_index", None),
    ("rollover", None),
    ("segments", _extract_args_index),
    ("shard_stores", _extract_args_index),
    ("shrink", _extract_args_index),
    ("simulate_index_template", None),
    ("simulate_template", None),
    ("split", _extract_args_index),
    ("stats", _extract_args_index),
    ("unfreeze", _extract_args_index),
    ("update_aliases", None),
    ("validate_query", _extract_args_index),
)


def instrument_elasticsearch_client_indices(module):
    # The module path was remapped in v8 to match previous versions.
    # In order to avoid double wrapping we check the version before
    # wrapping.
    if ES_VERSION < (8,):
        instrument_es_methods(module, "IndicesClient", _elasticsearch_client_indices_methods_below_v8, "indices")


def instrument_elasticsearch_client_indices_v8(module):
    instrument_es_methods(module, "IndicesClient", _elasticsearch_client_indices_methods_v8, "indices")


_elasticsearch_client_cat_methods_below_v8 = (
    ("aliases", None),
    ("allocation", None),
    ("count", _extract_args_index),
    ("fielddata", None),
    ("health", None),
    ("help", None),
    ("indices", _extract_args_index),
    ("master", None),
    ("nodes", None),
    ("pending_tasks", None),
    ("plugins", None),
    ("recovery", _extract_args_index),
    ("shards", _extract_args_index),
    ("segments", _extract_args_index),
    ("thread_pool", None),
)

_elasticsearch_client_cat_methods_v8 = (
    ("aliases", None),
    ("allocation", None),
    ("component_templates", None),
    ("count", _extract_args_index),
    ("fielddata", None),
    ("health", None),
    ("help", None),
    ("indices", _extract_args_index),
    ("master", None),
    ("ml_data_frame_analytics", None),
    ("ml_datafeeds", None),
    ("ml_jobs", None),
    ("ml_trained_models", None),
    ("nodeattrs", None),
    ("nodes", None),
    ("pending_tasks", None),
    ("plugins", None),
    ("recovery", _extract_args_index),
    ("repositories", None),
    ("segments", _extract_args_index),
    ("shards", _extract_args_index),
    ("snapshots", None),
    ("tasks", None),
    ("templates", None),
    ("thread_pool", None),
    ("transforms", None),
)


def instrument_elasticsearch_client_cat(module):
    # The module path was remapped in v8 to match previous versions.
    # In order to avoid double wrapping we check the version before
    # wrapping.
    if ES_VERSION < (8,):
        instrument_es_methods(module, "CatClient", _elasticsearch_client_cat_methods_below_v8, "cat")


def instrument_elasticsearch_client_cat_v8(module):
    instrument_es_methods(module, "CatClient", _elasticsearch_client_cat_methods_v8, "cat")


_elasticsearch_client_cluster_methods_below_v8 = (
    ("get_settings", None),
    ("health", _extract_args_index),
    ("pending_tasks", None),
    ("put_settings", None),
    ("reroute", None),
    ("state", _extract_args_metric_index),
    ("stats", None),
)


_elasticsearch_client_cluster_methods_v8 = (
    ("allocation_explain", _extract_args_allocation_explain_index),
    ("delete_component_template", None),
    ("delete_voting_config_exclusions", None),
    ("exists_component_template", None),
    ("get_component_template", None),
    ("get_settings", None),
    ("health", _extract_args_index),
    ("pending_tasks", None),
    ("post_voting_config_exclusions", None),
    ("put_component_template", None),
    ("put_settings", None),
    ("remote_info", None),
    ("reroute", None),
    ("state", _extract_args_metric_index),
    ("stats", None),
)


def instrument_elasticsearch_client_cluster(module):
    # The module path was remapped in v8 to match previous versions.
    # In order to avoid double wrapping we check the version before
    # wrapping.
    if ES_VERSION < (8,):
        instrument_es_methods(module, "ClusterClient", _elasticsearch_client_cluster_methods_below_v8, "cluster")


def instrument_elasticsearch_client_cluster_v8(module):
    instrument_es_methods(module, "ClusterClient", _elasticsearch_client_cluster_methods_v8, "cluster")


_elasticsearch_client_nodes_methods_below_v8 = (
    ("hot_threads", None),
    ("info", None),
    ("shutdown", None),
    ("stats", None),
)
_elasticsearch_client_nodes_methods_v8 = (
    ("clear_repositories_metering_archive", None),
    ("get_repositories_metering_info", None),
    ("hot_threads", None),
    ("info", None),
    ("reload_secure_settings", None),
    ("stats", None),
    ("usage", None),
)


def instrument_elasticsearch_client_nodes(module):
    # The module path was remapped in v8 to match previous versions.
    # In order to avoid double wrapping we check the version before
    # wrapping.
    if ES_VERSION < (8,):
        instrument_es_methods(module, "NodesClient", _elasticsearch_client_nodes_methods_below_v8, "nodes")


def instrument_elasticsearch_client_nodes_v8(module):
    instrument_es_methods(module, "NodesClient", _elasticsearch_client_nodes_methods_v8, "nodes")


_elasticsearch_client_snapshot_methods_below_v8 = (
    ("create", None),
    ("create_repository", None),
    ("delete", None),
    ("delete_repository", None),
    ("get", None),
    ("get_repository", None),
    ("restore", None),
    ("status", None),
    ("verify_repository", None),
)
_elasticsearch_client_snapshot_methods_v8 = (
    ("cleanup_repository", None),
    ("clone", None),
    ("create", None),
    ("create_repository", None),
    ("delete", None),
    ("delete_repository", None),
    ("get", None),
    ("get_repository", None),
    ("restore", None),
    ("status", None),
    ("verify_repository", None),
)


def instrument_elasticsearch_client_snapshot(module):
    # The module path was remapped in v8 to match previous versions.
    # In order to avoid double wrapping we check the version before
    # wrapping.
    if ES_VERSION < (8,):
        instrument_es_methods(module, "SnapshotClient", _elasticsearch_client_snapshot_methods_below_v8, "snapshot")


def instrument_elasticsearch_client_snapshot_v8(module):
    instrument_es_methods(module, "SnapshotClient", _elasticsearch_client_snapshot_methods_v8, "snapshot")


_elasticsearch_client_tasks_methods = (
    ("list", None),
    ("cancel", None),
    ("get", None),
)


def instrument_elasticsearch_client_tasks(module):
    # The module path was remapped in v8 to match previous versions.
    # In order to avoid double wrapping we check the version before
    # wrapping.
    if ES_VERSION < (8,):
        instrument_es_methods(module, "TasksClient", _elasticsearch_client_tasks_methods, "tasks")


def instrument_elasticsearch_client_tasks_v8(module):
    instrument_es_methods(module, "TasksClient", _elasticsearch_client_tasks_methods, "tasks")


_elasticsearch_client_ingest_methods_below_v8 = (
    ("get_pipeline", None),
    ("put_pipeline", None),
    ("delete_pipeline", None),
    ("simulate", None),
)

_elasticsearch_client_ingest_methods_v8 = (
    ("delete_pipeline", None),
    ("geo_ip_stats", None),
    ("get_pipeline", None),
    ("processor_grok", None),
    ("put_pipeline", None),
    ("simulate", None),
)


def instrument_elasticsearch_client_ingest(module):
    # The module path was remapped in v8 to match previous versions.
    # In order to avoid double wrapping we check the version before
    # wrapping.
    if ES_VERSION < (8,):
        instrument_es_methods(module, "IngestClient", _elasticsearch_client_ingest_methods_below_v8, "ingest")


def instrument_elasticsearch_client_ingest_v8(module):
    instrument_es_methods(module, "IngestClient", _elasticsearch_client_ingest_methods_v8, "ingest")


#
# Instrumentation to get Datastore Instance Information
#


def _nr_Connection__init__wrapper(wrapped, instance, args, kwargs):
    """Cache datastore instance info on Connection object"""

    def _bind_params(host="localhost", port=9200, *args, **kwargs):
        return host, port

    host, port = _bind_params(*args, **kwargs)
    port = str(port)
    instance._nr_host_port = (host, port)

    return wrapped(*args, **kwargs)


def instrument_elasticsearch_connection_base(module):
    wrap_function_wrapper(module, "Connection.__init__", _nr_Connection__init__wrapper)


def BaseNode__init__wrapper(wrapped, instance, args, kwargs):
    result = wrapped(*args, **kwargs)
    instance._nr_host_port = (instance.host, str(instance.port))
    return result


def instrument_elastic_transport__node__base(module):
    if hasattr(module, "BaseNode"):
        wrap_function_wrapper(module, "BaseNode.__init__", BaseNode__init__wrapper)


def _nr_get_connection_wrapper(wrapped, instance, args, kwargs):
    """Read instance info from Connection and stash on Transaction."""

    transaction = current_transaction()

    if transaction is None:
        return wrapped(*args, **kwargs)

    conn = wrapped(*args, **kwargs)

    instance_info = (None, None, None)
    try:
        tracer_settings = transaction.settings.datastore_tracer

        if tracer_settings.instance_reporting.enabled:
            host, port_path_or_id = conn._nr_host_port
            instance_info = (host, port_path_or_id, None)
    except Exception:
        instance_info = ("unknown", "unknown", None)

    transaction._nr_datastore_instance_info = instance_info

    return conn


def _nr_perform_request_wrapper(wrapped, instance, args, kwargs):
    """Read instance info from Connection and stash on Transaction."""

    transaction = current_transaction()

    if transaction is None:
        return wrapped(*args, **kwargs)

    if not hasattr(instance.node_pool.get, "_nr_wrapped"):
        instance.node_pool.get = function_wrapper(_nr_get_connection_wrapper)(instance.node_pool.get)
        instance.node_pool.get._nr_wrapped = True

    return wrapped(*args, **kwargs)


def instrument_elasticsearch_transport(module):
    if hasattr(module, "Transport") and hasattr(module.Transport, "get_connection"):
        wrap_function_wrapper(module, "Transport.get_connection", _nr_get_connection_wrapper)


def instrument_elastic_transport__transport(module):
    if hasattr(module, "Transport") and hasattr(module.Transport, "perform_request"):
        wrap_function_wrapper(module, "Transport.perform_request", _nr_perform_request_wrapper)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from conftest import cache_kafka_producer_headers
from testing_support.validators.validate_messagebroker_headers import (
    validate_messagebroker_headers,
)
from testing_support.validators.validate_transaction_errors import (
    validate_transaction_errors,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.background_task import background_task
from newrelic.common.object_names import callable_name
from newrelic.packages import six


def test_trace_metrics(topic, send_producer_message):
    from kafka.version import __version__ as version

    scoped_metrics = [("MessageBroker/Kafka/Topic/Produce/Named/%s" % topic, 1)]
    unscoped_metrics = scoped_metrics
    txn_name = "test_producer:test_trace_metrics.<locals>.test" if six.PY3 else "test_producer:test"

    @validate_transaction_metrics(
        txn_name,
        scoped_metrics=scoped_metrics,
        rollup_metrics=unscoped_metrics,
        custom_metrics=[("Python/MessageBroker/Kafka-Python/%s" % version, 1)],
        background_task=True,
    )
    @background_task()
    def test():
        send_producer_message()

    test()


def test_distributed_tracing_headers(topic, send_producer_message):
    txn_name = "test_producer:test_distributed_tracing_headers.<locals>.test" if six.PY3 else "test_producer:test"

    @validate_transaction_metrics(
        txn_name,
        rollup_metrics=[
            ("Supportability/TraceContext/Create/Success", 1),
            ("Supportability/DistributedTrace/CreatePayload/Success", 1),
        ],
        background_task=True,
    )
    @background_task()
    @cache_kafka_producer_headers
    @validate_messagebroker_headers
    def test():
        send_producer_message()

    test()


def test_producer_errors(topic, producer, monkeypatch):
    monkeypatch.setitem(producer.config, "value_serializer", None)
    monkeypatch.setitem(producer.config, "key_serializer", None)

    @validate_transaction_errors([callable_name(AssertionError)])
    @background_task()
    def test():
        with pytest.raises(AssertionError):
            producer.send(topic, value=object())
            producer.flush()

    test()
